"""L1 Norm."""

import torch

from mrpro.operators.Functional import ElementaryProximableFunctional, throw_if_negative_or_complex


class L1Norm(ElementaryProximableFunctional):
    r"""Functional class for the L1 Norm.

    This implements the functional given by
    :math:`f: C^N \rightarrow [0, \infty), x \rightarrow  \| W (x-b)\|_1`,
    where W is a either a scalar or tensor that corresponds to a (block-) diagonal operator
    that is applied to the input.

    In most cases, consider setting `divide_by_n` to `true` to be independent of input size.

    The norm of the vector is computed along the dimensions given at initialization.
    """

    def __call__(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        """Compute the L1 norm of the input tensor.

        Calculates :math:`|| W * (x - b) ||_1`, where :math:`W` is `weight` and :math:`b` is `target`.
        The norm is computed along dimensions specified by `dim`.
        If `divide_by_n` is true, the result is averaged over these dimensions;
        otherwise, it's summed.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            The L1 norm. If `keepdim` is true, the dimensions `dim` are retained
            with size 1; otherwise, they are reduced.
        """
        return super().__call__(x)

    def forward(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor]:
        """Apply forward of L1Norm.

        .. note::
            Prefer calling the instance of the L1Norm as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        value = (self.weight * (x - self.target)).abs()

        if self.divide_by_n:
            return (torch.mean(value, dim=self.dim, keepdim=self.keepdim),)
        else:
            return (torch.sum(value, dim=self.dim, keepdim=self.keepdim),)

    def prox(self, x: torch.Tensor, sigma: torch.Tensor | float = 1.0) -> tuple[torch.Tensor]:
        """Proximal Mapping of the L1 Norm.

        Compute the proximal mapping of the L1 norm.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal mapping applied to the input tensor
        """
        throw_if_negative_or_complex(sigma)
        diff = x - self.target
        threshold = self.weight * sigma
        threshold = self._divide_by_n(threshold, torch.broadcast_shapes(x.shape, threshold.shape))
        x_out = torch.sgn(diff) * torch.relu(diff.abs() - threshold.abs()) + self.target
        x_out = x_out.to(torch.result_type(threshold, x_out))
        return (x_out,)

    def prox_convex_conj(
        self,
        x: torch.Tensor,
        sigma: torch.Tensor | float = 1.0,
    ) -> tuple[torch.Tensor]:
        """Convex conjugate of the L1 Norm.

        Compute the proximal mapping of the convex conjugate of the L1 norm.

        Parameters
        ----------
        x
            input tensor
        sigma
            scaling factor

        Returns
        -------
            Proximal of the convex conjugate applied to the input tensor
        """
        throw_if_negative_or_complex(sigma)
        diff = x - sigma * self.target
        threshold = self._divide_by_n(self.weight.abs(), torch.broadcast_shapes(x.shape, self.weight.shape))
        x_out = torch.sgn(diff) * torch.clamp_max(diff.abs(), threshold.abs())
        x_out = x_out.to(torch.result_type(threshold, x_out))
        return (x_out,)



"""Class for Zero Pad Operator."""

from collections.abc import Sequence

import torch

from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils.pad_or_crop import pad_or_crop


class ZeroPadOp(LinearOperator):
    """Zero Pad operator class."""

    def __init__(self, dim: Sequence[int], original_shape: Sequence[int], padded_shape: Sequence[int]) -> None:
        """Zero Pad Operator class.

        The operator carries out zero-padding if the `padded_shape` is larger than `orig_shape` and cropping if the
        `padded_shape` is smaller.

        Parameters
        ----------
        dim
            dimensions along which padding should be applied
        original_shape
            shape of original data along dim, same length as `dim`
        padded_shape
            shape of padded data along dim, same length as `dim`
        """
        if len(dim) != len(original_shape) or len(dim) != len(padded_shape):
            raise ValueError('Dim, orig_shape and padded_shape have to be of same length')

        super().__init__()
        self.dim = dim
        self.original_shape = original_shape
        self.padded_shape = padded_shape

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply zero-padding or cropping to the input tensor.

        If `padded_shape` (defined at initialization) is larger than `original_shape`
        along the specified dimensions, the tensor is zero-padded. If smaller,
        it is cropped. The operation is applied along the dimensions specified in `dim`.

        Parameters
        ----------
        x
            Input tensor. Its shape along the dimensions in `dim` should
            match `original_shape`.

        Returns
        -------
            The padded or cropped tensor. Its shape along the dimensions in `dim`
            will match `padded_shape`.


        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of ZeroPadOp.

        .. note::
            Prefer calling the instance of the ZeroPadOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (pad_or_crop(x, self.padded_shape, self.dim),)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply cropping or zero-padding (adjoint of forward).

        This operation is the adjoint of the forward `ZeroPadOp`. If the forward
        operation was padding, the adjoint is cropping. If the forward was
        cropping, the adjoint is zero-padding. The operation is applied along
        the dimensions specified in `dim`.

        Parameters
        ----------
        x
            Input tensor. Its shape along the dimensions in `dim` should
            match `padded_shape` (from initialization).

        Returns
        -------
            The cropped or padded tensor. Its shape along the dimensions in `dim`
            will match `original_shape` (from initialization).
        """
        return (pad_or_crop(x, self.original_shape, self.dim),)



"""Optimizer Status base class."""

import torch
from typing_extensions import TypedDict


class OptimizerStatus(TypedDict):
    """Base class for OptimizerStatus."""

    solution: tuple[torch.Tensor, ...]
    """Current estimate(s) of the solution. """

    iteration_number: int
    """Current iteration of the (iterative) algorithm."""



"""Numerical Phantoms and Datasets"""

from mrpro.phantoms.EllipsePhantom import EllipsePhantom
from mrpro.phantoms.phantom_elements import EllipseParameters
from mrpro.phantoms import brainweb
from mrpro.phantoms.m4raw import M4RawDataset
from mrpro.phantoms import mdcnn
from mrpro.phantoms.fastmri import FastMRIKDataDataset, FastMRIImageDataset
from mrpro.phantoms import coils

__all__ = [
    "EllipseParameters",
    "EllipsePhantom",
    "FastMRIImageDataset",
    "FastMRIKDataDataset",
    "M4RawDataset",
    "brainweb",
    "coils",
    "mdcnn"
]


"""Conjugate Gradient for linear systems with self-adjoint positive semidefinite linear operator."""

from collections.abc import Callable, Sequence

import torch
from typing_extensions import Unpack, overload

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.LinearOperatorMatrix import LinearOperatorMatrix


def vdot(a: Sequence[torch.Tensor], b: Sequence[torch.Tensor]) -> torch.Tensor:
    """Vector dot product."""
    return sum(
        (torch.vdot(a_i.flatten(), b_i.flatten()) for a_i, b_i in zip(a, b, strict=True)), start=torch.tensor(0.0)
    )


class CGStatus(OptimizerStatus):
    """Status of the conjugate gradient algorithm."""

    residual: tuple[torch.Tensor, ...]
    """Residual of the current estimate."""

    preconditioned_residual: tuple[torch.Tensor, ...]
    """Preconditioned residual of the current estimate."""


OperatorMatrixLikeCallable = Callable[[Unpack[tuple[torch.Tensor, ...]]], tuple[torch.Tensor, ...]]
OperatorLikeCallable = Callable[[torch.Tensor], tuple[torch.Tensor]]


@overload
def cg(
    operator: OperatorLikeCallable,
    right_hand_side: tuple[torch.Tensor] | torch.Tensor,
    *,
    initial_value: tuple[torch.Tensor] | torch.Tensor | None = None,
    preconditioner_inverse: OperatorLikeCallable | None = None,
    max_iterations: int = 128,
    tolerance: float = 1e-4,
    callback: Callable[[CGStatus], bool | None] | None = None,
) -> tuple[torch.Tensor]: ...


@overload
def cg(
    operator: OperatorMatrixLikeCallable,
    right_hand_side: Sequence[torch.Tensor],
    *,
    initial_value: Sequence[torch.Tensor] | None = None,
    preconditioner_inverse: OperatorMatrixLikeCallable | None = None,
    max_iterations: int = 128,
    tolerance: float = 1e-4,
    callback: Callable[[CGStatus], bool | None] | None = None,
) -> tuple[torch.Tensor, ...]: ...


def cg(
    operator: LinearOperator | LinearOperatorMatrix | OperatorMatrixLikeCallable | OperatorLikeCallable,
    right_hand_side: Sequence[torch.Tensor] | torch.Tensor,
    *,
    initial_value: Sequence[torch.Tensor] | torch.Tensor | None = None,
    preconditioner_inverse: LinearOperator
    | LinearOperatorMatrix
    | OperatorMatrixLikeCallable
    | OperatorLikeCallable
    | None = None,
    max_iterations: int = 128,
    tolerance: float = 1e-4,
    callback: Callable[[CGStatus], bool | None] | None = None,
) -> tuple[torch.Tensor, ...] | tuple[torch.Tensor]:
    r"""(Preconditioned) Conjugate Gradient for solving :math:`Hx=b`.

    This algorithm solves systems of the form :math:`H x = b`, where :math:`H` is a self-adjoint positive semidefinite
    linear operator and :math:`b` is the right-hand side.

     The method performs the following steps:

     1. Initialize the residual :math:`r_0 = b - Hx_0` (with :math:`x_0` as the initial guess).
     2. Set the search direction :math:`p_0 = r_0`.
     3. For each iteration :math:`k = 0, 1, 2, ...`:

        - Compute :math:`\alpha_k = \frac{r_k^T r_k}{p_k^T H p_k}`.
        - Update the solution: :math:`x_{k+1} = x_k + \alpha_k p_k`.
        - Compute the new residual: :math:`r_{k+1} = r_k - \alpha_k H p_k`.
        - Update the search direction: :math:`\beta_k = \frac{r_{k+1}^T r_{k+1}}{r_k^T r_k}`,
            then :math:`p_{k+1} = r_{k+1} + \beta_k p_k`.

    The operator can be either a `LinearOperator` or a `LinearOperatorMatrix`.
    In both cases, this implementation does not verify if the operator is self-adjoint and positive semidefinite.
    It will silently return the wrong result if the assumptions are not met.
    It can solve a batch of :math:`N` systems jointly if `right_hand_side` has a batch dimension
    and the operator interprets the batch dimension as :math:`H` being block-diagonal with blocks :math:`H_i`
    resulting in :math:`b = [b_1, ..., b_N] ^T`.

    If `preconditioner_inverse` is provided, it solves :math:`M^{-1}Hx = M^{-1}b`
    implicitly, where `preconditioner_inverse(r)` computes :math:`M^{-1}r`.

    See [Hestenes1952]_, [Nocedal2006]_, and [WikipediaCG]_ for more information.


    Parameters
    ----------
    operator
        Self-adjoint operator :math:`H`.
    right_hand_side
        Right-hand-side :math:`b`.
    initial_value
        Initial guess :math:`x_0`. If `None`, the initial guess is set to the zero vector.
    preconditioner_inverse
        Preconditioner :math:`M^{-1}`. If None, no preconditioning is applied.
    max_iterations
        Maximum number of iterations.
    tolerance
        Tolerance for the L2 norm of the residual :math:`\|r_k\|_2`.
    callback
        Function called at each iteration. See `CGStatus`.

    Returns
    -------
        An approximate solution of the linear system :math:`Hx=b`.

    References
    ----------
    .. [Hestenes1952] Hestenes, M. R., & Stiefel, E. (1952). Methods of conjugate gradients for solving linear systems.
       Journal of Research of the National Bureau of Standards , 49(6), 409-436
    .. [Nocedal2006] Nocedal, J. (2006). *Numerical Optimization* (2nd ed.). Springer.
    .. [WikipediaCG] Wikipedia: Conjugate Gradient https://en.wikipedia.org/wiki/Conjugate_gradient
    """
    right_hand_side_ = (right_hand_side,) if isinstance(right_hand_side, torch.Tensor) else tuple(right_hand_side)
    if initial_value is None:  # start with zero initial value
        solution = tuple(torch.zeros_like(r) for r in right_hand_side_)
        residual = right_hand_side_
    else:  # start with provided initial value
        solution = (initial_value,) if isinstance(initial_value, torch.Tensor) else tuple(initial_value)
        if len(solution) != len(right_hand_side_):
            raise ValueError('Length mismatch in initial_value and right_hand_side')
        if any(s.shape != r.shape for s, r in zip(solution, right_hand_side_, strict=True)):
            raise ValueError('Shape mismatch in initial_value and right_hand_side')
        residual = tuple(rhs - op_sol for rhs, op_sol in zip(right_hand_side_, operator(*solution), strict=True))

    if preconditioner_inverse is not None:
        conjugate = preconditioner_inverse(*residual)
    else:
        conjugate = residual

    # dummy value. new value will be set in loop before first usage
    direction_dot_residual_old: torch.Tensor | None = None

    for iteration in range(max_iterations):
        residual_dot_residual = vdot(residual, residual).real

        if residual_dot_residual < tolerance**2:  # are we done?
            break

        if preconditioner_inverse is not None:
            direction = preconditioner_inverse(*residual)
            direction_dot_residual = vdot(residual, direction).real
        else:
            direction = residual
            direction_dot_residual = residual_dot_residual

        if direction_dot_residual_old is not None:  # not first iteration
            beta = direction_dot_residual / direction_dot_residual_old
            conjugate = tuple(d + beta * con for d, con in zip(direction, conjugate, strict=True))
        operator_conjugate = operator(*conjugate)
        alpha = direction_dot_residual / vdot(conjugate, operator_conjugate).real
        solution = tuple(sol + alpha * con for sol, con in zip(solution, conjugate, strict=True))
        residual = tuple(res - alpha * op_con for res, op_con in zip(residual, operator_conjugate, strict=True))
        direction_dot_residual_old = direction_dot_residual

        if callback is not None:
            continue_iterations = callback(
                {
                    'solution': solution,
                    'iteration_number': iteration,
                    'residual': residual,
                    'preconditioned_residual': direction,
                }
            )
            if continue_iterations is False:
                break

    return solution



"""Numerical phantom with ellipses."""

from collections.abc import Sequence

import torch
from einops import repeat

from mrpro.data.KData import KData
from mrpro.data.KHeader import KHeader
from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.phantoms.phantom_elements import EllipseParameters


class EllipsePhantom:
    """Numerical phantom as the sum of different ellipses."""

    def __init__(self, ellipses: Sequence[EllipseParameters] | None = None):
        """Initialize ellipse phantom.

        Parameters
        ----------
        ellipses
            Parameters defining the ellipses.
            If `None`, defaults to three ellipses with different parameters.
        """
        if ellipses is None:
            self.ellipses = [
                EllipseParameters(center_x=0.2, center_y=0.2, radius_x=0.1, radius_y=0.25, intensity=1),
                EllipseParameters(center_x=0.1, center_y=-0.1, radius_x=0.3, radius_y=0.1, intensity=2),
                EllipseParameters(center_x=-0.2, center_y=0.2, radius_x=0.18, radius_y=0.25, intensity=4),
            ]
        else:
            self.ellipses = list(ellipses)

    def kspace(self, ky: torch.Tensor, kx: torch.Tensor) -> torch.Tensor:
        """Create 2D analytic kspace data based on given k-space locations.

        For a corresponding image with 256 x 256 voxels, the k-space locations should be defined within ``[-128, 127]``.

        The Fourier representation of ellipses can be analytically described by Bessel functions [KOA2007]_.

        Parameters
        ----------
        ky
            k-space locations in ky (phase encoding direction).
        kx
            k-space locations in kx (frequency encoding direction).

        Returns
        -------
            K-space data.

        References
        ----------
        .. [KOA2007] Koay C, Sarlls J, Oezarslan E (2007) Three-dimensional analytical magnetic resonance imaging
           phantom in the Fourier domain. MRM 58(2) https://doi.org/10.1002/mrm.21292
        ..
        """
        # kx and ky should be broadcastable
        try:
            shape = torch.broadcast_shapes(kx.shape, ky.shape)
        except RuntimeError:
            raise ValueError(f'shape mismatch between kx {kx.shape} and ky {ky.shape}') from None
        if kx.device != ky.device:
            raise ValueError(f'device mismatch between kx {kx.device} and ky {ky.device}')

        kdata = torch.zeros(shape, dtype=torch.complex64, device=kx.device)
        for ellipse in self.ellipses:
            arg = torch.sqrt((ellipse.radius_x * 2) ** 2 * kx**2 + (ellipse.radius_y * 2) ** 2 * ky**2)
            arg[arg < 1e-6] = 1e-6  # avoid zeros

            cdata = 2 * 2 * ellipse.radius_x * ellipse.radius_y * 0.5 * torch.special.bessel_j1(torch.pi * arg) / arg
            kdata = kdata + (
                torch.exp(-1j * 2 * torch.pi * (ellipse.center_x * kx + ellipse.center_y * ky))
                * cdata
                * ellipse.intensity
            )

        # Scale k-space data by factor sqrt(number of points) to ensure correct scaling after FFT with
        # normalization "ortho". See e.g. https://docs.scipy.org/doc/scipy/tutorial/fft.html
        kdata = kdata * kdata.numel() ** 0.5
        return kdata

    def image_space(self, image_dimensions: SpatialDimension[int]) -> torch.Tensor:
        """Create image representation of phantom.

        Parameters
        ----------
        image_dimensions
            Number of voxels in the image.
            This is a 2D simulation, so the output will be of shape `(1 1 1 image_dimensions.y image_dimensions.x)`.

        Returns
        -------
            Image representation of phantom
        """
        # Calculate image representation of phantom
        ny, nx = image_dimensions.y, image_dimensions.x
        ix, iy = torch.meshgrid(
            torch.linspace(-nx // 2, nx // 2 - 1, nx),
            torch.linspace(-ny // 2, ny // 2 - 1, ny),
            indexing='xy',
        )

        idata = torch.zeros((ny, nx), dtype=torch.complex64)
        for ellipse in self.ellipses:
            in_ellipse = (
                (ix / nx - ellipse.center_x) ** 2 / ellipse.radius_x**2
                + (iy / ny - ellipse.center_y) ** 2 / ellipse.radius_y**2
            ) <= 1
            idata += ellipse.intensity * in_ellipse

        return repeat(idata, 'y x->other coils z y x', other=1, coils=1, z=1)

    def kdata(self, trajectory: KTrajectory, encoding_matrix: SpatialDimension[int]) -> KData:
        """Create k-space data for the phantom.

        Parameters
        ----------
        trajectory
            Trajectory.
        encoding_matrix
            Encoding matrix.

        Returns
        -------
            K-space data with header and trajectory.
        """
        if (trajectory.kz != 0).any():
            raise ValueError('Only 2D k-space data is supported')

        data = self.kspace(trajectory.ky, trajectory.kx)
        header = KHeader(
            recon_fov=SpatialDimension(z=0.01, y=1, x=1),
            encoding_fov=SpatialDimension(z=0.01, y=1, x=1),
            encoding_matrix=encoding_matrix,
            recon_matrix=encoding_matrix,
        )
        kdata = KData(data=data, header=header, traj=trajectory)
        return kdata



"""Cartesian Sampling Operator."""

import warnings

import torch
from einops import rearrange, repeat
from typing_extensions import Self

from mrpro.data.enums import TrajType
from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils.reduce_repeat import reduce_repeat
from mrpro.utils.reshape import unsqueeze_left


class CartesianSamplingOp(LinearOperator):
    """Cartesian Sampling Operator.

    Selects points on a Cartesian grid based on the k-space trajectory.
    Thus, the adjoint sorts the data into regular Cartesian sampled grid based on the k-space
    trajectory. Non-acquired points are zero-filled.
    """

    def __init__(self, encoding_matrix: SpatialDimension[int], traj: KTrajectory) -> None:
        """Initialize Sampling Operator class.

        Parameters
        ----------
        encoding_matrix
            shape of the encoded k-space.
            Only values for directions in which the trajectory is Cartesian will be used
            in the adjoint to determine the shape after reordering,
            i.e., the operator's domain.
        traj
            the k-space trajectory describing at which frequencies data is sampled.
            Its broadcasted shape will be used to determine the shape after sampling,
            i.e., the operator's range.
        """
        super().__init__()
        # the shape of the k data,
        sorted_grid_shape = SpatialDimension.from_xyz(encoding_matrix)

        # Cache as these might take some time to compute
        traj_type_kzyx = traj.type_along_kzyx
        traj_device = traj.device
        ktraj_tensor = traj.as_tensor()

        # If a dimension is irregular or singleton, we will not perform any reordering
        # in it and the shape of data will remain.
        # only dimensions on a cartesian grid will be reordered.
        # The device of the input trajectory is matched.
        if traj_type_kzyx[-1] == TrajType.ONGRID:  # kx
            kx_idx = ktraj_tensor[-1, ...].round().to(dtype=torch.int64) + sorted_grid_shape.x // 2
        else:
            sorted_grid_shape.x = ktraj_tensor.shape[-1]
            kx_idx = repeat(
                torch.arange(ktraj_tensor.shape[-1], device=traj_device),
                'k0->other coils k2 k1 k0',
                other=1,
                coils=1,
                k2=1,
                k1=1,
            )

        if traj_type_kzyx[-2] == TrajType.ONGRID:  # ky
            ky_idx = ktraj_tensor[-2, ...].round().to(dtype=torch.int64) + sorted_grid_shape.y // 2
        else:
            sorted_grid_shape.y = ktraj_tensor.shape[-2]
            ky_idx = repeat(
                torch.arange(ktraj_tensor.shape[-2], device=traj_device),
                'k1->other coils k2 k1 k0',
                other=1,
                coils=1,
                k2=1,
                k0=1,
            )

        if traj_type_kzyx[-3] == TrajType.ONGRID:  # kz
            kz_idx = ktraj_tensor[-3, ...].round().to(dtype=torch.int64) + sorted_grid_shape.z // 2
        else:
            sorted_grid_shape.z = ktraj_tensor.shape[-3]
            kz_idx = repeat(
                torch.arange(ktraj_tensor.shape[-3], device=traj_device),
                'k2->other coils k2 k1 k0',
                other=1,
                coils=1,
                k1=1,
                k0=1,
            )

        # 1D indices into a flattened tensor.
        kidx = kz_idx * sorted_grid_shape.y * sorted_grid_shape.x + ky_idx * sorted_grid_shape.x + kx_idx
        kidx = rearrange(kidx, '... kz ky kx -> ... (kz ky kx)')

        # check that all points are inside the encoding matrix
        inside_encoding_matrix = (
            ((kx_idx >= 0) & (kx_idx < sorted_grid_shape.x))
            & ((ky_idx >= 0) & (ky_idx < sorted_grid_shape.y))
            & ((kz_idx >= 0) & (kz_idx < sorted_grid_shape.z))
        )
        if not torch.all(inside_encoding_matrix):
            warnings.warn(
                'K-space points lie outside of the encoding_matrix and will be ignored.'
                ' Increase the encoding_matrix to include these points.',
                stacklevel=2,
            )

            inside_encoding_matrix = rearrange(inside_encoding_matrix, '... kz ky kx -> ... (kz ky kx)')
            inside_encoding_matrix_idx = inside_encoding_matrix.nonzero(as_tuple=True)[-1]
            inside_encoding_matrix_idx = torch.reshape(inside_encoding_matrix_idx, (*kidx.shape[:-1], -1))
            self._inside_encoding_matrix_idx: torch.Tensor | None = inside_encoding_matrix_idx
            kidx = torch.take_along_dim(kidx, inside_encoding_matrix_idx, dim=-1)
        else:
            self._inside_encoding_matrix_idx = None

        self._fft_idx = kidx

        # we can skip the indexing if the data is already sorted
        self._needs_indexing = (
            not torch.all(torch.diff(kidx) == 1)
            or traj.shape[-3:] != sorted_grid_shape.zyx
            or self._inside_encoding_matrix_idx is not None
        )

        self._trajectory_shape = traj.shape
        self._sorted_grid_shape = sorted_grid_shape

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Forward operator which selects acquired k-space data from k-space.

        Parameters
        ----------
        x
            k-space, fully sampled (or zerofilled) and sorted in Cartesian dimensions
            with shape given by encoding_matrix.

        Returns
        -------
            Selected k-space data in acquired shape (as described by the trajectory).
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of CartesianSamplingOp.

        .. note::
            Prefer calling the instance of the CartesianSamplingOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self._sorted_grid_shape != SpatialDimension(*x.shape[-3:]):
            raise ValueError('k-space data shape mismatch')

        if not self._needs_indexing:
            return (x,)

        x_kflat = rearrange(x, '... coils k2_enc k1_enc k0_enc -> ... coils (k2_enc k1_enc k0_enc)')
        # take_along_dim broadcasts, but needs the same number of dimensions
        idx = unsqueeze_left(self._fft_idx, x_kflat.ndim - self._fft_idx.ndim)
        x_inside_encoding_matrix = torch.take_along_dim(x_kflat, idx, dim=-1)

        if self._inside_encoding_matrix_idx is None:
            # all trajectory points are inside the encoding matrix
            x_indexed = x_inside_encoding_matrix
        else:
            # we need to add zeros
            x_indexed = self._broadcast_and_scatter_along_last_dim(
                x_inside_encoding_matrix,
                self._trajectory_shape[-1] * self._trajectory_shape[-2] * self._trajectory_shape[-3],
                self._inside_encoding_matrix_idx,
            )

        # reshape to (... other coils, k2, k1, k0)
        x_reshaped = x_indexed.reshape(x.shape[:-3] + self._trajectory_shape[-3:])

        return (x_reshaped,)

    def adjoint(self, y: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint operator sorting data into the encoding_space matrix.

        Parameters
        ----------
        y
            k-space data in acquired shape.

        Returns
        -------
            k-space data sorted into encoding_space matrix. Non-acquired points are zero-filled.
        """
        if self._trajectory_shape[-3:] != y.shape[-3:]:
            raise ValueError('k-space data shape mismatch')

        if not self._needs_indexing:
            return (y,)

        y_kflat = rearrange(y, '... coils k2 k1 k0 -> ... coils (k2 k1 k0)')

        if self._inside_encoding_matrix_idx is not None:
            idx = unsqueeze_left(self._inside_encoding_matrix_idx, y_kflat.ndim - self._inside_encoding_matrix_idx.ndim)
            y_kflat = torch.take_along_dim(y_kflat, idx, dim=-1)

        y_scattered = self._broadcast_and_scatter_along_last_dim(
            y_kflat, self._sorted_grid_shape.z * self._sorted_grid_shape.y * self._sorted_grid_shape.x, self._fft_idx
        )

        # reshape to  ..., other, coils, k2_enc, k1_enc, k0_enc
        y_reshaped = y_scattered.reshape(
            *y.shape[:-3],
            self._sorted_grid_shape.z,
            self._sorted_grid_shape.y,
            self._sorted_grid_shape.x,
        )

        return (y_reshaped,)

    @staticmethod
    def _broadcast_and_scatter_along_last_dim(
        data_to_scatter: torch.Tensor, n_last_dim: int, scatter_index: torch.Tensor
    ) -> torch.Tensor:
        """Broadcast scatter index and scatter into zero tensor.

        Parameters
        ----------
        data_to_scatter
            Data to be scattered at indices scatter_index
        n_last_dim
            Number of data points in last dimension
        scatter_index
            Indices describing where to scatter data

        Returns
        -------
            Data scattered into tensor along scatter_index
        """
        # scatter does not broadcast, so we need to manually broadcast the indices
        broadcast_shape = torch.broadcast_shapes(scatter_index.shape[:-1], data_to_scatter.shape[:-1])
        idx_expanded = torch.broadcast_to(scatter_index, (*broadcast_shape, scatter_index.shape[-1]))

        # although scatter_ is inplace, this will not cause issues with autograd, as self
        # is always constant zero and gradients w.r.t. src work as expected.
        data_scattered = torch.zeros(
            *broadcast_shape,
            n_last_dim,
            dtype=data_to_scatter.dtype,
            device=data_to_scatter.device,
        ).scatter_(dim=-1, index=idx_expanded, src=data_to_scatter)

        return data_scattered

    @property
    def gram(self) -> 'CartesianMaskingOp':
        """Return the Gram operator for this Cartesian Sampling Operator.

        Returns
        -------
            Gram operator for this Cartesian Sampling Operator.
        """
        return CartesianMaskingOp.from_sampling_op(self)

    def __repr__(self) -> str:
        """Representation method for CartesianSamplingOperator."""
        device = self._fft_idx.device if self._fft_idx is not None else 'none'
        if self._inside_encoding_matrix_idx is None:
            enc_matrix_warning = ''
        else:
            enc_matrix_warning = (
                '\nk-space points lie outside of the encoding_matrix and will be ignored.'
                '\nIncrease the encoding_matrix to include these points.'
            )

        out = (
            f'{type(self).__name__} on device: {device}\n'
            f'Needs indexing: {self._needs_indexing}\n'
            f'Sorted grid shape: {self._sorted_grid_shape}'
            f'{enc_matrix_warning}'
        )
        return out


class CartesianMaskingOp(LinearOperator):
    """Cartesian Masking Operator.

    The Cartesian Masking Operator is the composition `CartesianSamplingOp.H @ CartesianSamplingOp`,
    which sets to zero all non sampled Cartesian k-space points.
    """

    def __init__(self, mask: torch.Tensor | None):
        """Initialize Cartesian Sampling Masking Operator from a mask.

        Parameters
        ----------
        mask
            The mask to use for the Cartesian Masking Operator.

        """
        super().__init__()
        self.mask = None if mask is None else reduce_repeat(mask.float())

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the Gram operator (Cartesian Masking).

        This operator applies a mask to the input tensor, effectively zeroing out
        points that were not sampled according to the Cartesian trajectory.
        It represents `CartesianSamplingOp.H @ CartesianSamplingOp`.

        Parameters
        ----------
        x
            Input k-space data, typically fully sampled or zero-filled.

        Returns
        -------
            Masked k-space data.
        """
        return super().__call__(x)

    @classmethod
    def from_trajectory(cls, traj: KTrajectory, encoding_matrix: SpatialDimension[int]) -> 'CartesianMaskingOp':
        """Initialize Cartesian Sampling Masking Operator from a trajectory.

        Parameters
        ----------
        traj
            The trajectory to use for the Cartesian Masking Operator.
        encoding_matrix
            The encoding matrix to use for the Cartesian Masking Operator.
        """
        return cls.from_sampling_op(CartesianSamplingOp(encoding_matrix, traj))

    @classmethod
    def from_sampling_op(cls, sampling_op: CartesianSamplingOp) -> Self:
        """Initialize Cartesian Sampling Masking Operator from a Cartesian Sampling Operator.

        Parameters
        ----------
        sampling_op
            The Cartesian Sampling Operator for which to create the Gram operator.
        """
        if sampling_op._needs_indexing:
            ones = torch.ones(
                *sampling_op._trajectory_shape[:-3],
                *sampling_op._sorted_grid_shape.zyx,
                device=sampling_op._fft_idx.device,
            )
            (mask,) = sampling_op.adjoint(*sampling_op.forward(ones))
        else:
            mask = None
        return cls(mask)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of CartesianMaskingOp.

        .. note::
            Prefer calling the instance of the CartesianMaskingOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.mask is None:
            return (x,)
        return (x * self.mask,)

    def adjoint(self, y: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the Gram operator (Cartesian Masking).

        Since the Cartesian Masking operator is self-adjoint (it involves
        applying a mask, which is a real-valued multiplication), its adjoint
        is the same as its forward operation.

        Parameters
        ----------
        y
            Input k-space data.

        Returns
        -------
            Masked k-space data.
        """
        return self.forward(y)

    @property
    def H(self) -> Self:  # noqa: N802
        """Return the adjoint of the Cartesian Masking Operator.

        Returns
        -------
            the same operator, as the masking operator is self-adjoint.
        """
        return self



"""Spoiled gradient echo signal model."""

import torch

from mrpro.operators.SignalModel import SignalModel
from mrpro.utils.reshape import unsqueeze_right


class SpoiledGRE(SignalModel[torch.Tensor, torch.Tensor, torch.Tensor]):
    """Spoiled gradient echo signal model."""

    def __init__(
        self, flip_angle: float | torch.Tensor, echo_time: float | torch.Tensor, repetition_time: float | torch.Tensor
    ) -> None:
        r"""Initialize spoiled gradient echo signal model.

        Assumes perfect spoiling and a longitudinal steady state.
        This is a simplified special case of `~mrpro.operators.models.TransientSteadyStateWithPreparation`.

        The model is defined as:
        :math:`S = M_0 e^{-t_r / T_2^*}  \frac{\sin(\alpha)(1 - e^{-t_e / T_1})}{(1 - \cos(\alpha) e^{-t_e / T_1})}`

        where :math:`M_0` is the equilibrium magnetization, :math:`\alpha` is the flip angle,
        :math:`t_e` is the echo time, and :math:`t_r` is the repetition time.

        Parameters
        ----------
        flip_angle
            Flip angle in radians.
        echo_time
            Echo time.
        repetition_time
            Repetition time.
        """
        super().__init__()
        self.flip_angle = torch.nn.Parameter(torch.as_tensor(flip_angle))
        self.echo_time = torch.nn.Parameter(torch.as_tensor(echo_time))
        self.repetition_time = torch.nn.Parameter(torch.as_tensor(repetition_time))

    def __call__(self, m0: torch.Tensor, t1: torch.Tensor, t2star: torch.Tensor) -> tuple[torch.Tensor,]:
        """Calculate the Spoiled Gradient Echo signal.

        Parameters
        ----------
        m0
            Equilibrium signal / proton density.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t1
            Longitudinal (T1) relaxation time.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.
        t2star
            Effective transverse (T2*) relaxation time.
            Shape `(...)`, for example `(*other, coils, z, y, x)` or `(samples)`.

        Returns
        -------
            Signal
            Shape `(1 ...)`, for example `(1, *other, coils, z, y, x)` or `(1, samples)`.
        """
        return super().__call__(m0, t1, t2star)

    def forward(self, m0: torch.Tensor, t1: torch.Tensor, t2star: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of SpoiledGRE.

        .. note::
            Prefer calling the instance of the SpoiledGRE as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ndim = max(m0.ndim, t1.ndim, t2star.ndim) + 1
        flip_angle = unsqueeze_right(self.flip_angle, ndim - self.flip_angle.ndim)
        echo_time = unsqueeze_right(self.echo_time, ndim - self.echo_time.ndim)
        repetition_time = unsqueeze_right(self.repetition_time, ndim - self.repetition_time.ndim)
        e1 = torch.exp(-repetition_time / t1)
        e2 = torch.exp(-echo_time / t2star)
        signal = m0 * torch.sin(flip_angle) * (1 - e1) * e2 / (1 - torch.cos(flip_angle) * e1)
        return (signal,)



"""Random generator."""

from collections.abc import Sequence
from math import ceil, floor

import torch


def check_bounds(low: float | int | torch.Tensor, high: float | int | torch.Tensor, dtype: torch.dtype | None) -> None:
    """Clip the bounds to a range matching the dtype.

    Parameters
    ----------
    low
        Lower bound.
    high
        Upper bound.
    dtype
        Data type, used to find allowed range.
    """
    info: torch.finfo | torch.iinfo
    if dtype is None:
        info = torch.finfo()
        minval, maxval = info.min, info.max
    elif dtype.is_floating_point:
        info = torch.finfo(dtype)
        minval, maxval = info.min, info.max
    else:
        info = torch.iinfo(dtype)
        minval = info.min
        if dtype in (torch.int64, torch.uint64):
            maxval = info.max  # https://github.com/pytorch/pytorch/issues/81446
        else:
            maxval = info.max + 1
    if low > high:
        raise ValueError('low should be lower than high')
    if low < minval or high > maxval:
        raise ValueError(f'low/high should be in the range of {info.min} and {info.max} for {dtype}')


class RandomGenerator:
    """Generate random numbers for various purposes.

    Uses a fixed seed to ensure reproducibility.

    Provides:
        - Scalar uniform random numbers:
            int8, int16, int32, int64, uint8, uint16, uint32, uint64,
            float32, float64, complex64, complex128
        - Tensor of uniform random numbers:
            int8_tensor, int16_tensor, int32_tensor, int64_tensor, uint8_tensor,
            float32_tensor, float64_tensor, complex64_tensor, complex128_tensor
            (Note: uint16, uint32, uint64 tensors are not yet supported by PyTorch)
        - Tuple of uniform random numbers:
            int8_tuple, int16_tuple, int32_tuple, int64_tuple,
            uint8_tuple, uint16_tuple, uint32_tuple, uint64_tuple,
            float32_tuple, float64_tuple, complex64_tuple, complex128_tuple
    """

    def __init__(self, seed: int | None = None):
        """Initialize the random generator with a fixed seed.

        Parameters
        ----------
        seed
            Seed for the random generator. If `None`, default torch seeding is used.
        """
        self.generator = torch.Generator()
        if seed is not None:
            self.generator.manual_seed(seed)

    def _randint(
        self,
        size: Sequence[int] | int,
        low: int,
        high: int,
        dtype: torch.dtype = torch.int64,
        device: torch.device | None = None,
    ) -> torch.Tensor:
        """Generate uniform random integers in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).
        dtype
            Data type of the output tensor.
        device
            Device of the output tensor. If `None`, the tensor is created on the default device.

        Returns
        -------
            Tensor of random integers.
        """
        check_bounds(low, high, dtype)
        size_ = (size,) if isinstance(size, int) else size
        return torch.randint(low, high, size_, generator=self.generator, dtype=dtype, device=device)

    def _rand(
        self,
        size: Sequence[int] | int,
        low: float | torch.Tensor,
        high: float | torch.Tensor,
        dtype: torch.dtype = torch.float32,
        device: torch.device | None = None,
    ) -> torch.Tensor:
        """Generate uniform random floats in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound.
        high
            Upper bound.
        dtype
            Data type of the output tensor.
        device
            Device of the output tensor. If `None`, the tensor is created on the default device.

        Returns
        -------
            Tensor of random floats.
        """
        check_bounds(low, high, dtype)
        return (torch.rand(size, generator=self.generator, dtype=dtype, device=device) * (high - low)) + low

    def float32_tensor(self, size: Sequence[int] | int = (1,), low: float = 0.0, high: float = 1.0) -> torch.Tensor:
        """Generate a float32 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Tensor of float32 random numbers.
        """
        return self._rand(size, low, high, torch.float32)

    def float64_tensor(self, size: Sequence[int] | int = (1,), low: float = 0.0, high: float = 1.0) -> torch.Tensor:
        """Generate a float64 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Tensor of float64 random numbers.
        """
        return self._rand(size, low, high, torch.float64)

    def complex64_tensor(self, size: Sequence[int] | int = (1,), low: float = 0.0, high: float = 1.0) -> torch.Tensor:
        """Generate a complex64 tensor with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Tensor of complex64 random numbers.
        """
        if low < 0:
            raise ValueError('low/high refer to the amplitude and must be positive')
        amp = self.float32_tensor(size, low, high)
        phase = self.float32_tensor(size, -torch.pi, torch.pi)
        return (amp * torch.exp(1j * phase)).to(dtype=torch.complex64)

    def complex128_tensor(self, size: Sequence[int] | int = (1,), low: float = 0.0, high: float = 1.0) -> torch.Tensor:
        """Generate a complex128 tensor with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Tensor of complex128 random numbers.
        """
        if low < 0:
            raise ValueError('low/high refer to the amplitude and must be positive')
        amp = self.float64_tensor(size, low, high)
        phase = self.float64_tensor(size, -torch.pi, torch.pi)
        return (amp * torch.exp(1j * phase)).to(dtype=torch.complex128)

    def int8_tensor(self, size: Sequence[int] | int = (1,), low: int = -1 << 7, high: int = 1 << 7) -> torch.Tensor:
        """Generate an int8 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tensor of int8 random numbers.
        """
        return self._randint(size, low, high, dtype=torch.int8)

    def int16_tensor(self, size: Sequence[int] | int = (1,), low: int = -1 << 15, high: int = 1 << 15) -> torch.Tensor:
        """Generate an int16 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tensor of int16 random numbers.
        """
        return self._randint(size, low, high, dtype=torch.int16)

    def int32_tensor(self, size: Sequence[int] | int = (1,), low: int = -1 << 31, high: int = 1 << 31) -> torch.Tensor:
        """Generate an int32 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tensor of int32 random numbers.
        """
        return self._randint(size, low, high, dtype=torch.int32)

    def int64_tensor(
        self, size: Sequence[int] | int = (1,), low: int = -1 << 63, high: int = (1 << 63) - 1
    ) -> torch.Tensor:
        """Generate an int64 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).
            Maximum value is (1 << 63) - 1 due to https://github.com/pytorch/pytorch/issues/81446

        Returns
        -------
            Tensor of int64 random numbers.
        """
        return self._randint(size, low, high, dtype=torch.int64)

    def uint8_tensor(self, size: Sequence[int] | int = (1,), low: int = 0, high: int = 1 << 8) -> torch.Tensor:
        """Generate a uint8 tensor with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tensor of uint8 random numbers.
        """
        return self._randint(size, low, high, dtype=torch.uint8)

    def bool_tensor(self, size: Sequence[int] | int = (1,)) -> torch.Tensor:
        """Generate boolean tensor of given size.

        Parameters
        ----------
        size
            Shape of the output tensor.
        """
        return self.uint8_tensor(size, low=0, high=2).bool()

    def bool(self) -> bool:
        """Generate a random boolean value.

        Returns
        -------
            Random boolean.
        """
        return self.uint8(0, 1) == 1

    def float32(self, low: float = 0.0, high: float = 1.0) -> float:
        """Generate a float32 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Random float32 number.
        """
        return self.float32_tensor((1,), low, high).item()

    def float64(self, low: float = 0.0, high: float = 1.0) -> float:
        """Generate a float64 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Random float64 number.
        """
        return self.float64_tensor((1,), low, high).item()

    def complex64(self, low: float = 0, high: float = 1.0) -> complex:
        """Generate a complex64 scalar with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Random complex64 number.
        """
        return self.complex64_tensor((1,), low, high).item()

    def complex128(self, low: float = 0, high: float = 1.0) -> complex:
        """Generate a complex128 scalar with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Random complex128 number.
        """
        return self.complex128_tensor((1,), low, high).item()

    def uint8(self, low: int = 0, high: int = (1 << 8) - 1) -> int:
        """Generate a uint8 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random uint8 number.
        """
        return int(self.uint8_tensor((1,), low, high).item())

    def uint16(self, low: int = 0, high: int = 1 << 16) -> int:
        """Generate a uint16 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random uint16 number.
        """
        if low < 0 or high > 1 << 16:
            raise ValueError('Low must be positive and high must be <= 2^16')
        return int(self.int32_tensor((1,), low, high).item())

    def uint32(self, low: int = 0, high: int = 1 << 32) -> int:
        """Generate a uint32 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random uint32 number.
        """
        if low < 0 or high > 1 << 32:
            raise ValueError('Low must be positive and high must be <= 2^32')
        return int(self.int64_tensor((1,), low, high).item())

    def int8(self, low: int = -1 << 7, high: int = 1 << 7) -> int:
        """Generate an int8 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random int8 number.
        """
        return int(self.int8_tensor((1,), low, high).item())

    def int16(self, low: int = -1 << 15, high: int = 1 << 15) -> int:
        """Generate an int16 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random int16 number.
        """
        return int(self.int16_tensor((1,), low, high).item())

    def int32(self, low: int = -1 << 31, high: int = 1 << 31) -> int:
        """Generate an int32 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random int32 number.
        """
        return int(self.int32_tensor((1,), low, high).item())

    def int64(self, low: int = -1 << 63, high: int = (1 << 63) - 1) -> int:
        """Generate an int64 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random int64 number.
        """
        return int(self.int64_tensor((1,), low, high).item())

    def uint64(self, low: int = 0, high: int = (1 << 64) - 1) -> int:
        """Generate a uint64 scalar with uniform distribution in [low, high).

        Parameters
        ----------
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Random uint64 number.
        """
        if low < 0 or high > 1 << 64:
            raise ValueError('Low must be positive and high must be <= 2^64')
        range_ = high - low
        new_low = -1 << 63
        new_high = new_low + range_
        value = self.int64(new_low, new_high) - new_low + low
        return value

    def float32_tuple(self, size: int, low: float = 0, high: float = 1) -> tuple[float, ...]:
        """Generate a tuple of float32 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Tuple of float32 random numbers.
        """
        return tuple(self.float32_tensor((size,), low, high))

    def float64_tuple(self, size: int, low: float = 0, high: float = 1) -> tuple[float, ...]:
        """Generate a tuple of float64 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Tuple of float64 random numbers.
        """
        return tuple(self.float64_tensor((size,), low, high))

    def complex64_tuple(self, size: int, low: float = 0, high: float = 1) -> tuple[complex, ...]:
        """Generate a tuple of complex64 numbers with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Tuple of complex64 random numbers.
        """
        return tuple(self.complex64_tensor((size,), low, high))

    def complex128_tuple(self, size: int, low: float = 0, high: float = 1) -> tuple[complex, ...]:
        """Generate a tuple of complex128 numbers with uniform amplitude in [low, high).

        The phase is uniformly distributed in [-π, π].

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound for amplitude (must be non-negative).
        high
            Upper bound for amplitude.

        Returns
        -------
            Tuple of complex128 random numbers.
        """
        return tuple(self.complex128_tensor((size,), low, high))

    def uint8_tuple(self, size: int, low: int = 0, high: int = 1 << 8) -> tuple[int, ...]:
        """Generate a tuple of uint8 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of uint8 random numbers.
        """
        return tuple(self.uint8_tensor((size,), low, high))

    def uint16_tuple(self, size: int, low: int = 0, high: int = 1 << 16) -> tuple[int, ...]:
        """Generate a tuple of uint16 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of uint16 random numbers.
        """
        return tuple([self.uint16(low, high) for _ in range(size)])

    def uint32_tuple(self, size: int, low: int = 0, high: int = 1 << 32) -> tuple[int, ...]:
        """Generate a tuple of uint32 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of uint32 random numbers.
        """
        return tuple([self.uint32(low, high) for _ in range(size)])

    def uint64_tuple(self, size: int, low: int = 0, high: int = (1 << 64) - 1) -> tuple[int, ...]:
        """Generate a tuple of uint64 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of uint64 random numbers.
        """
        return tuple([self.uint64(low, high) for _ in range(size)])

    def int8_tuple(self, size: int, low: int = -1 << 7, high: int = 1 << 7) -> tuple[int, ...]:
        """Generate a tuple of int8 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of int8 random numbers.
        """
        return tuple(self.int8_tensor((size,), low, high))

    def int16_tuple(self, size: int, low: int = -1 << 15, high: int = 1 << 15) -> tuple[int, ...]:
        """Generate a tuple of int16 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of int16 random numbers.
        """
        return tuple(self.int16_tensor((size,), low, high))

    def int32_tuple(self, size: int, low: int = -1 << 31, high: int = 1 << 31) -> tuple[int, ...]:
        """Generate a tuple of int32 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of int32 random numbers.
        """
        return tuple(self.int32_tensor((size,), low, high))

    def int64_tuple(self, size: int, low: int = -1 << 63, high: int = (1 << 63) - 1) -> tuple[int, ...]:
        """Generate a tuple of int64 numbers with uniform distribution in [low, high).

        Parameters
        ----------
        size
            Number of elements.
        low
            Lower bound (inclusive).
        high
            Upper bound (exclusive).

        Returns
        -------
            Tuple of int64 random numbers.
        """
        return tuple(self.int64_tensor((size,), low, high))

    def ascii(self, size: int) -> str:
        """Generate a random ASCII string.

        Parameters
        ----------
        size
            Length of the string.

        Returns
        -------
            Random ASCII string.
        """
        return ''.join([chr(self.uint8(32, 127)) for _ in range(size)])

    def rand_like(self, x: torch.Tensor, low: float = 0.0, high: float = 1.0) -> torch.Tensor:
        """Generate a tensor with the same shape, device, and dtype as `x`, filled with uniform random numbers.

        Parameters
        ----------
        x
            Reference tensor.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Random tensor with the same shape and dtype as `x`.
        """
        return self.rand_tensor(x.shape, x.dtype, low=low, high=high).to(x.device)

    def randn_like(self, x: torch.Tensor) -> torch.Tensor:
        """Generate a tensor with the same shape, device, and dtype as `x`, filled with standard normal random numbers.

        Parameters
        ----------
        x
            Reference tensor.

        Returns
        -------
            Random tensor with the same shape and dtype as `x`.
        """
        return self.randn_tensor(x.shape, x.dtype).to(x.device)

    def rand_tensor(
        self,
        size: Sequence[int],
        dtype: torch.dtype,
        low: float | int = 0,
        high: int | float = 1,
    ) -> torch.Tensor:
        """Generate a tensor of given shape and dtype with uniform random numbers in [low, high).

        Parameters
        ----------
        size
            Shape of the output tensor.
        dtype
            Data type of the output tensor.
        low
            Lower bound.
        high
            Upper bound.

        Returns
        -------
            Random tensor.
        """
        if dtype.is_complex:
            tensor = self.complex64_tensor(size, low, high).to(dtype=dtype)
        elif dtype.is_floating_point:
            tensor = self._rand(size, low, high, dtype)
        elif dtype == torch.bool:
            tensor = self.int32_tensor(size, low=0, high=1) > 0
        else:
            tensor = self._randint(size, ceil(low), floor(high), dtype)
        return tensor

    def randn_tensor(self, size: Sequence[int], dtype: torch.dtype) -> torch.Tensor:
        """Generate a tensor of given shape and dtype with standard normal distribution.

        Parameters
        ----------
        size
            Shape of the output tensor.
        dtype
            Data type of the output tensor.

        Returns
        -------
            Random tensor with normal distribution.
        """
        return torch.randn(size=size, generator=self.generator, dtype=dtype)

    def randperm(self, n: int, *, dtype: torch.dtype = torch.int64) -> torch.Tensor:
        """Generate a random permutation of integers from 0 to n - 1.

        Parameters
        ----------
        n
            Number of elements.
        dtype
            Data type of the output tensor.

        Returns
        -------
            Tensor containing a random permutation.
        """
        return torch.randperm(n, generator=self.generator, dtype=dtype)

    def gaussian_variable_density_samples(
        self, shape: Sequence[int], low: int, high: int, fwhm: float = float('inf'), always_sample: Sequence[int] = ()
    ) -> torch.Tensor:
        """Generate Gaussian variable density samples.

        Generates indices in [low, high[ with a gaussian weighting.

        Parameters
        ----------
        shape
            Shape of the output tensor. The generated indices are 1D and in the last dimension.
            All other dimensions are batch dimensions.
        low
            Lower bound of the sampling domain.
        high
            Upper bound of the sampling domain.
        fwhm
            Full-width at half-maximum of the Gaussian.
        always_sample
            indices that should always included in the samples.
            For example, `range(-n_center//2, n_center//2)`

        Returns
        -------
            1D tensor of selected indices.
        """
        *n_batch, n_samples = shape
        if n_samples > high - low:
            raise ValueError('n_samples must be <= (high - low)')
        n_random = n_samples - len(always_sample)
        if n_random < 0:
            raise ValueError('more always sampled points requested than total number of samples')
        elif n_random == 0:
            return torch.tensor(always_sample).sort().values.broadcast_to(*n_batch, -1)
        pdf = torch.exp(-torch.tensor(2.0).log() * (2 * torch.arange(low, high) / fwhm) ** 2)
        pdf[[x - low for x in always_sample]] = 0
        if len(shape) > 1:
            pdf = pdf.broadcast_to((*n_batch, -1)).flatten(end_dim=-2)

        idx_rand = pdf.multinomial(n_random, False, generator=self.generator) + low
        if len(shape) > 1:
            idx_rand = idx_rand.unflatten(0, n_batch)
        idx_always = torch.tensor(always_sample).broadcast_to(*n_batch, -1)
        return torch.cat([idx_rand, idx_always], -1).sort().values



"""1D, 2D and 3D density compensation function calculation with voronoi method."""

from concurrent.futures import ProcessPoolExecutor
from itertools import product

import numpy as np
import torch
from numpy.typing import ArrayLike
from scipy.spatial import ConvexHull, Voronoi

UNIQUE_ROUNDING_DECIMALS = 15


def _volume(v: ArrayLike):
    return ConvexHull(v).volume


def dcf_1d(traj: torch.Tensor) -> torch.Tensor:
    """Calculate sample density compensation function for 1D trajectory.

    This function operates on a single `other` sample.
    See also `~mrpro.data.DcfData` and `~mrpro.utils.smap`.

    Parameters
    ----------
    traj
        k-space positions, 1D tensor

    Returns
    -------
        density compensation values
    """
    traj_sorted, inverse, counts = torch.unique(
        torch.round(traj, decimals=UNIQUE_ROUNDING_DECIMALS),
        sorted=True,
        return_inverse=True,
        return_counts=True,
    )

    # For a sorted trajectory: x0 x1 x2 ... xN
    # we assign the point at x1 the area (x1 - x0) / 2  + (x2 - x1) / 2,
    # this is done by central differences (-1,0,1).
    # For the edges, we append/prepend the values afterwards, such that:
    # we assign the point at x0 the area (x1 - x0) / 2 + (x1 - x0) / 2, and
    # we assign the point at xN the area (xN - xN-1) / 2 + (xN - xN-1) / 2.

    kernel = torch.tensor([-1 / 2, 0, 1 / 2], dtype=torch.float32, device=traj.device).reshape(1, 1, 3)

    if (elements := len(traj_sorted)) >= 3:
        central_diff = torch.nn.functional.conv1d(traj_sorted[None, None, :], kernel)[0, 0]
        first = traj_sorted[1] - traj_sorted[0]
        last = traj_sorted[-1] - traj_sorted[-2]
        central_diff = torch.cat((first[None], central_diff, last[None]), -1)
    elif elements == 2:
        diff = traj_sorted[1] - traj_sorted[0]
        central_diff = torch.cat((diff[None], diff[None]), -1)
    else:
        central_diff = torch.ones_like(traj_sorted)

    # Repeated points are reduced by the number of repeats
    dcf = torch.nan_to_num(central_diff / counts)[inverse]
    return dcf


def dcf_2d3d_voronoi(traj: torch.Tensor) -> torch.Tensor:
    """Calculate sample density compensation function using Voronoi method.

    This function computes the DCF by determining the area around each point in k-space using the Voronoi tessellation.
    Points at the edge of k-space are detected as outliers and are assigned the area of the 1% largest DCF values.

    The Voronoi tessellation assigns each point in k-space a region based on the proximity to its nearest neighbors. The
    DCF is then computed based on the inverse of the area of these regions.

    This function operates on a single `other` sample.
    See also `~mrpro.data.DcfData` and `~mrpro.utils.smap`.

    Parameters
    ----------
    traj
        k-space positions `(2 or 3, k2, k1, k0)`

    Returns
    -------
        density compensation values `(1, k2, k1, k0)`
    """
    # 2D and 3D trajectories supported
    dim = traj.shape[0]
    if dim not in (2, 3):
        raise ValueError(f'Only 2D or 3D trajectories supported, not {dim}D.')

    # Calculate dcf only for unique k-space positions
    traj_dim = traj.shape
    traj_numpy = np.round(traj.cpu().numpy(), decimals=UNIQUE_ROUNDING_DECIMALS)
    traj_numpy = traj_numpy.reshape(dim, -1)
    traj_unique, inverse, counts = np.unique(traj_numpy, return_inverse=True, return_counts=True, axis=1)

    # Especially in 3D, errors in the calculation of the convex hull can occur for edge points. To avoid this,
    # the corner points of a cube bounding box are added here. The bounding box is chosen very large to ensure these
    # edge points of the trajectory can still be accurately detected in the outlier detection further down.
    furthest_corner = np.max(np.abs(traj_unique))
    corner_points = np.array(list(product([-1, 1], repeat=dim))) * furthest_corner * 10
    traj_extendend = np.concatenate((traj_unique, corner_points.transpose()), axis=1)

    # Carry out voronoi tessellation
    vdiagram = Voronoi(traj_extendend.transpose())
    regions = [vdiagram.regions[r] for r in vdiagram.point_region[: -len(corner_points)]]  # Ignore corner points
    vertices = [vdiagram.vertices[region] for region in regions]

    if dim == 2:
        # Shoelace equation for 2d
        def cross2d(x: np.ndarray, y: np.ndarray) -> np.ndarray:
            return x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]

        dcf = np.array([np.abs(cross2d(v[:-1], v[1:]).sum(0) + cross2d(v[-1], v[0])) / 2 for v in vertices])

    else:
        # Calculate volume/area of voronoi cells using processes, as this is a very time-consuming operation
        # and ConvexHull is singlethreaded and does not seem to drop the GIL
        # TODO: this could maybe be made faster as the polyhedrons are known to be convex
        future = ProcessPoolExecutor(max_workers=torch.get_num_threads()).map(_volume, vertices, chunksize=100)
        dcf = np.array(list(future))

    # Get outliers (i.e. voronoi cell which are unbound) and set them to a reasonable value
    # Outliers are defined as values larger than 1.5 * inter quartile range of the values
    # Outliers are set to the average of the 1% largest values.
    dcf_sorted = np.sort(dcf)
    q1, q3 = np.percentile(dcf_sorted, [25, 75])
    iqr = q3 - q1
    upper_bound = q3 + 1.5 * iqr
    idx_outliers = np.nonzero(dcf > upper_bound)
    n_outliers = len(idx_outliers[0])
    high_values_start = int(0.99 * (len(dcf_sorted) - n_outliers))
    high_values_end = len(dcf_sorted) - n_outliers  # this works also for n_outliers==0
    fill_value = np.average(dcf_sorted[high_values_start:high_values_end])
    dcf[idx_outliers] = fill_value

    # Sort dcf values back into the original order (i.e. before calling unique)
    dcf = np.reshape((dcf / counts)[inverse], traj_dim[1:])

    return torch.tensor(dcf, dtype=torch.float32, device=traj.device)



"""Generalized Sum Multiplication Operator."""

import re

import torch
from einops import einsum

from mrpro.operators.LinearOperator import LinearOperator


class EinsumOp(LinearOperator):
    r"""A Linear Operator that implements sum products in Einstein notation.

    Implements :math:`A_{\mathrm{indices}_A}*x^{\mathrm{indices}_x} = y_{\mathrm{indices}_y}`
    with Einstein summation rules over the :math:`indices`, see `torch.einsum` or `einops.einsum`
    for more information. Note, that the indices must be space separated (einops convention).


    It can be used to implement tensor contractions, such as for example, different versions of
    matrix-vector or matrix-matrix products of the form `A @ x`, depending on the chosen einsum rules and
    shapes of `A` and `x`.

    Examples are:

    - matrix-vector multiplication of :math:`A` and the batched vector :math:`x = [x_1, ..., x_N]` consisting
      of :math:`N` vectors :math:`x_1, x_2, ..., x_N`. Then, the operation defined by
      :math:`A @ x := \mathrm{diag}(A, A, ..., A) [x_1, x_2, ..., x_N]^T` = :math:`[A x_1, A x_2, ..., A x_N]^T`
      can be implemented by the einsum rule ``'i j, ... j -> ... i'``.

    - matrix-vector multiplication of a matrix :math:`A` consisting of :math:`N` different matrices
      :math:`A_1, A_2, ... A_N` with one vector :math:`x`. Then, the operation defined by
      :math:`A @ x := \mathrm{diag}(A_1, A_2,..., A_N) [x, x, ..., x]^T`
      can be implemented by the einsum rule ``'... i j, j -> ... i'``.

    - matrix-vector multiplication of a matrix :math:`A` consisting of :math:`N` different matrices
      :math:`A_1, A_2, ... A_N` with a vector :math:`x = [x_1,...,x_N]` consisting
      of :math:`N` vectors :math:`x_1, x_2, ..., x_N`. Then, the operation defined by
      :math:`A @ x := \mathrm{diag}(A_1, A_2,..., A_N) [x_1, x_2, ..., x_N]^T`
      can be implemented by the einsum rule ``'... i j, ... j -> ... i'``.
      This is the default behavior of the operator.
    """

    def __init__(self, matrix: torch.Tensor, einsum_rule: str = '... i j, ... j -> ... i') -> None:
        """Initialize Einsum Operator.

        Parameters
        ----------
        matrix
            Matrix :math:`A` to be used as first factor in the sum product :math:`A*x`

        einsum_rule
            Einstein summation rule describing the forward of the operator.
            Also see torch.einsum for more information.
        """
        super().__init__()
        if (match := re.match('(.+),(.+)->(.+)', einsum_rule)) is None:
            raise ValueError(f'Einsum pattern should match (.+),(.+)->(.+) but got {einsum_rule}.')
        indices_matrix, indices_input, indices_output = match.groups()
        # swapping the input and output indices gets the adjoint rule
        self._adjoint_pattern = f'{indices_matrix},{indices_output}->{indices_input}'
        self._forward_pattern = einsum_rule
        self.matrix = torch.nn.Parameter(matrix, matrix.requires_grad)

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply sum-product of input `x` with the operator's matrix `A`.

        :math:`A` and the rule used to perform the sum-product is set at initialization.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the sum-product operation.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of EinsumOp.

        .. note::
            Prefer calling the instance of the EinsumOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        y = einsum(self.matrix, x, self._forward_pattern)
        return (y,)

    def adjoint(self, y: torch.Tensor) -> tuple[torch.Tensor]:
        """Multiplication of input with the adjoint of :math:`A`.

        Parameters
        ----------
        y
            Tensor to be multiplied with hermitian/adjoint matrix :math:`A`

        Returns
        -------
            Result of the adjoint sum-product operation.
        """
        x = einsum(self.matrix.conj(), y, self._adjoint_pattern)
        return (x,)



"""Linear operators (such as FourierOp), functionals/loss functions, and qMRI signal models."""

from mrpro.operators.Operator import Operator
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.Functional import FunctionalType, ProximableFunctional, ElementaryFunctional, ElementaryProximableFunctional, ScaledProximableFunctional
from mrpro.operators import functionals, models
from mrpro.operators.AveragingOp import AveragingOp
from mrpro.operators.CartesianSamplingOp import CartesianSamplingOp, CartesianMaskingOp
from mrpro.operators.ConjugateGradientOp import ConjugateGradientOp
from mrpro.operators.ConstraintsOp import ConstraintsOp
from mrpro.operators.DensityCompensationOp import DensityCompensationOp
from mrpro.operators.DictionaryMatchOp import DictionaryMatchOp
from mrpro.operators.EinsumOp import EinsumOp
from mrpro.operators.FastFourierOp import FastFourierOp
from mrpro.operators.FiniteDifferenceOp import FiniteDifferenceOp
from mrpro.operators.FourierOp import FourierOp
from mrpro.operators.GridSamplingOp import GridSamplingOp
from mrpro.operators.IdentityOp import IdentityOp
from mrpro.operators.Jacobian import Jacobian
from mrpro.operators.LinearOperatorMatrix import LinearOperatorMatrix
from mrpro.operators.MagnitudeOp import MagnitudeOp
from mrpro.operators.MultiIdentityOp import MultiIdentityOp
from mrpro.operators.NonUniformFastFourierOp import NonUniformFastFourierOp
from mrpro.operators.OptimizerOp import OptimizerOp
from mrpro.operators.PatchOp import PatchOp
from mrpro.operators.PCACompressionOp import PCACompressionOp
from mrpro.operators.PhaseOp import PhaseOp
from mrpro.operators.ProximableFunctionalSeparableSum import ProximableFunctionalSeparableSum
from mrpro.operators.RearrangeOp import RearrangeOp
from mrpro.operators.SensitivityOp import SensitivityOp
from mrpro.operators.SignalModel import SignalModel
from mrpro.operators.SliceProjectionOp import SliceProjectionOp
from mrpro.operators.WaveletOp import WaveletOp
from mrpro.operators.ZeroPadOp import ZeroPadOp
from mrpro.operators.ZeroOp import ZeroOp


__all__ = [
    "AveragingOp",
    "CartesianMaskingOp",
    "CartesianSamplingOp",
    "ConjugateGradientOp",
    "ConstraintsOp",
    "DensityCompensationOp",
    "DictionaryMatchOp",
    "EinsumOp",
    "ElementaryFunctional",
    "ElementaryProximableFunctional",
    "FastFourierOp",
    "FiniteDifferenceOp",
    "FourierOp",
    "FunctionalType",
    "GridSamplingOp",
    "IdentityOp",
    "Jacobian",
    "LinearOperator",
    "LinearOperatorMatrix",
    "MagnitudeOp",
    "MultiIdentityOp",
    "NonUniformFastFourierOp",
    "Operator",
    "OptimizerOp",
    "PCACompressionOp",
    "PatchOp",
    "PhaseOp",
    "ProximableFunctional",
    "ProximableFunctionalSeparableSum",
    "RearrangeOp",
    "ScaledProximableFunctional",
    "SensitivityOp",
    "SignalModel",
    "SliceProjectionOp",
    "WaveletOp",
    "ZeroOp",
    "ZeroPadOp",
    "functionals",
    "models"
]


"""Identity Operator."""

import torch

from mrpro.operators.LinearOperator import LinearOperator


class IdentityOp(LinearOperator):
    r"""The Identity Operator.

    A Linear Operator that returns a single input unchanged.
    """

    def __init__(self) -> None:
        """Initialize Identity Operator."""
        super().__init__()

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the identity operation.

        This operator returns the input tensor unchanged.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            The input tensor, unchanged.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of IdentityOp.

        .. note::
            Prefer calling the instance of the IdentityOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (x,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply the adjoint of the identity operation.

        Since the identity operator is self-adjoint, this method returns
        the input tensor unchanged.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            The input tensor, unchanged.
        """
        return (x,)



"""Tensor reshaping utilities."""

from collections.abc import Sequence
from functools import lru_cache
from math import prod

import einops
import torch

from mrpro.utils.typing import endomorph


def unsqueeze_right(x: torch.Tensor, n: int) -> torch.Tensor:
    """Unsqueeze multiple times in the rightmost dimension.

    Example:
        Tensor with shape `(1, 2, 3)` and `n=2` would result in a tensor with shape `(1, 2, 3, 1, 1)`.

    Parameters
    ----------
    x
        Tensor to unsqueeze.
    n
        Number of times to unsqueeze.

    Returns
    -------
        Unsqueezed tensor (view).
    """
    new_shape = list(x.shape) + [1] * n
    return x.reshape(new_shape)


def unsqueeze_left(x: torch.Tensor, n: int) -> torch.Tensor:
    """Unsqueeze multiple times in the leftmost dimension.

    Example:
        Tensor with shape `(1, 2, 3)` and `n=2` would result in a tensor with shape `(1, 1, 1, 2, 3)`.

    Parameters
    ----------
    x
        Tensor to unsqueeze.
    n
        Number of times to unsqueeze.

    Returns
    -------
        Unsqueezed tensor (view).
    """
    new_shape = [1] * n + list(x.shape)
    return x.reshape(new_shape)


def unsqueeze_at(x: torch.Tensor, dim: int, n: int) -> torch.Tensor:
    """Unsqueeze multiple times at a specific dimension.

    Example:
        Tensor with shape `(1, 2, 3)`, `dim=2`, and `n=2` would result in a tensor with shape `(1, 2, 1, 1, 3)`.

    Parameters
    ----------
    x
        Tensor to unsqueeze.
    dim
        Dimension to unsqueeze. Negative values are allowed.
    n
        Number of times to unsqueeze.
    """
    if n == 0:
        return x
    elif n == 1:
        return x.unsqueeze(dim)
    elif n < 0:
        raise ValueError('n must be positive')
    if not (-x.ndim - 1 <= dim <= x.ndim):
        raise IndexError(f'Dimension {dim} out of range for tensor of dimension {x.ndim}')
    if dim < 0:
        # dim=-1 should index after the last axis, etc., to match unsqueeze
        dim = x.ndim + dim + 1
    return x.reshape(*x.shape[:dim], *(n * (1,)), *x.shape[dim:])


@endomorph
def unsqueeze_tensors_at(*x, dim: int, ndim: int | None = None) -> tuple[torch.Tensor, ...]:
    """Unsqueeze tensors at a specific dimension to the same number of dimensions.

    Example:
        - Tensors with shapes `(1, 2, 3)` and `(1, 3)` and `dim=-2`
          result in tensors with shapes `(1, 2, 3)` and `(1, 1, 3)`, as the maximum number
          of input dimensions is 3.
        - Tensors with shapes `(1, 2, 3)` and `(1, 3)` and `dim=1` and `ndim=4`
          result in tensors with shapes `(1, 1, 2, 3)` and `(1, 1, 1, 3)`.

    Parameters
    ----------
    x
        Tensors to unsqueeze.
    dim
        Dimension to unsqueeze.
    ndim
        Number of dimensions to unsqueeze to. If `None`, unsqueeze to the maximum number of dimensions
        of the input tensors.

    Returns
    -------
        Unsqueezed tensors (views) with the same number of dimensions.
    """
    if ndim is None:
        ndim_ = max(el.ndim for el in x)
    elif ndim < min(el.ndim for el in x):
        raise ValueError('ndim must be greater or equal to the minimum number of dimensions of the input tensors')
    else:
        ndim_ = ndim
    return tuple(unsqueeze_at(el, dim, n=ndim_ - el.ndim) for el in x)


@endomorph
def unsqueeze_tensors_left(*x: torch.Tensor, ndim: int | None = None) -> tuple[torch.Tensor, ...]:
    """Unsqueeze tensors on the left to the same number of dimensions.

    Parameters
    ----------
    x
        Tensors to unsqueeze.
    ndim
        Minimum number of dimensions to unsqueeze to. If `None`, unsqueeze to the maximum number of dimensions
        of the input tensors.

    Returns
    -------
        Unsqueezed tensors (views).
    """
    ndim_ = max(el.ndim for el in x)
    if ndim is not None:
        ndim_ = max(ndim_, ndim)
    return tuple(unsqueeze_left(el, ndim_ - el.ndim) for el in x)


@endomorph
def unsqueeze_tensors_right(*x: torch.Tensor, ndim: int | None = None) -> tuple[torch.Tensor, ...]:
    """Unsqueeze tensors on the right to the same number of dimensions.

    Parameters
    ----------
    x
        Tensors to unsqueeze.
    ndim
        Minimum number of dimensions to unsqueeze to. If `None`, unsqueeze to the maximum number of dimensions
        of the input tensors.

    Returns
    -------
        Unsqueezed tensors (views).
    """
    ndim_ = max(el.ndim for el in x)
    if ndim is not None:
        ndim_ = max(ndim_, ndim)
    return tuple(unsqueeze_right(el, ndim_ - el.ndim) for el in x)


@endomorph
def broadcast_right(*x: torch.Tensor) -> tuple[torch.Tensor, ...]:
    """Broadcasting on the right.

    Given multiple tensors, apply broadcasting with unsqueezed on the right.
    First, tensors are unsqueezed on the right to the same number of dimensions.
    Then, `torch.broadcast_tensors` is used.

    ```{note}
    `broadcast_left` is regular `torch.broadcast_tensors`.
    ```

    Example:
        Tensors with shapes `(1, 2, 3)`, `(1, 2)`, `(2)` result in tensors with shapes `(2, 2, 3)`.

    Parameters
    ----------
    x
        Tensors to broadcast.

    Returns
    -------
        Broadcasted tensors (views).
    """
    max_dim = max(el.ndim for el in x)
    unsqueezed = torch.broadcast_tensors(*(unsqueeze_right(el, max_dim - el.ndim) for el in x))
    return unsqueezed


def reduce_view(x: torch.Tensor, dim: int | Sequence[int] | None = None) -> torch.Tensor:
    """Reduce expanded dimensions in a view to singletons.

    Reduce either all or specific dimensions to a singleton if it
    points to the same memory address.
    This undoes `torch.Tensor.expand`.

    Parameters
    ----------
    x
        Input tensor.
    dim
        Only reduce expanded dimensions in the specified dimensions.
        If `None`, reduce all expanded dimensions.
    """
    if dim is None:
        dim_: Sequence[int] = range(x.ndim)
    elif isinstance(dim, Sequence):
        dim_ = [d % x.ndim for d in dim]
    else:
        dim_ = [dim % x.ndim]

    stride = x.stride()
    newsize = [
        1 if stride == 0 and d in dim_ else oldsize
        for d, (oldsize, stride) in enumerate(zip(x.size(), stride, strict=True))
    ]
    return torch.as_strided(x, newsize, stride)


@lru_cache
def _reshape_idx(
    old_shape: tuple[int, ...], new_shape: tuple[int, ...], old_stride: tuple[int, ...]
) -> tuple[slice, ...]:
    """Get reshape reduce index (cached helper function for `reshape_broadcasted`).

    This function tries to group axes from new_shape and old_shape into the smallest groups that have
    the same number of elements, starting from the right.
    If all axes of old shape of a group are stride=0 dimensions, we can reduce them.

    Example:
        old_shape = (30, 2, 2, 3)
        new_shape = `(6, 5, 4, 3)`
        Will result in the groups (starting from the right):
            - old: 3     new: 3
            - old: 2, 2  new: 4
            - old: 30    new: 6, 5
        Only the "old" groups are important.
        If all axes that are grouped together in an "old" group are stride 0 (=broadcasted),
        we can collapse them to singleton dimensions.

    This function returns the indexer that either collapses dimensions to singleton or keeps all
    elements, i.e., the slices in the returned list are all either slice(1) or slice(None).
    """
    idx = []
    pointer_old, pointer_new = len(old_shape) - 1, len(new_shape) - 1  # start from the right
    while pointer_old >= 0:
        product_new, product_old = 1, 1  # the number of elements in the current "new" and "old" group
        group: list[int] = []
        while product_old != product_new or not group:
            if product_old <= product_new:
                # increase "old" group
                product_old *= old_shape[pointer_old]
                group.append(pointer_old)
                pointer_old -= 1
            else:
                # increase "new" group
                # we don't need to track the new group, just the number of elements covered.
                product_new *= new_shape[pointer_new]
                pointer_new -= 1
        # we found a group. now we need to decide what to do.
        if all(old_stride[d] == 0 for d in group):
            # all dimensions are broadcasted
            # -> reduce to singleton
            idx.extend([slice(1)] * len(group))
        else:
            # preserve dimension
            idx.extend([slice(None)] * len(group))
    idx = idx[::-1]  # we worked right to left, but our index should be left to right
    return tuple(idx)


def reshape_broadcasted(tensor: torch.Tensor, *shape: int) -> torch.Tensor:
    """Reshape a tensor while preserving broadcasted (stride 0) dimensions where possible.

    Parameters
    ----------
    tensor
        The input tensor to reshape.
    shape
        The target shape for the tensor. One of the values can be ``-1`` and its size will be inferred.

    Returns
    -------
        A tensor reshaped to the target shape, preserving broadcasted dimensions where feasible.
    """
    try:
        # if we can view the tensor directly, it will preserve broadcasting
        return tensor.view(shape)
    except RuntimeError:
        # we cannot do a view, we need to do more work:

        # -1 means infer size, i.e., the remaining elements of the input not already covered by the other axes.
        negative_ones = shape.count(-1)
        size = tensor.shape.numel()
        if not negative_ones:
            if prod(shape) != size:
                # use same exception as pytorch
                raise RuntimeError(f"shape '{list(shape)}' is invalid for input of size {size}") from None
        elif negative_ones > 1:
            raise RuntimeError('only one dimension can be inferred') from None
        elif negative_ones == 1:
            # we need to figure out the size of the "-1" dimension
            known_size = -prod(shape)  # negative, as it includes the -1
            if size % known_size:
                # non-integer result. no possible size of the -1 axis exists.
                raise RuntimeError(f"shape '{list(shape)}' is invalid for input of size {size}") from None
            shape = tuple(size // known_size if s == -1 else s for s in shape)

        # most of the broadcasted dimensions can be preserved: only dimensions that are joined with non-
        # broadcasted dimensions cannot be preserved and must be made contiguous.
        # all dimensions that can be preserved as broadcasted are first collapsed to singleton,
        # such that contiguous does not create copies along these axes.
        idx = _reshape_idx(tensor.shape, shape, tensor.stride())
        # make contiguous only in dimensions in which broadcasting cannot be preserved
        semicontiguous = tensor[idx].contiguous()
        # finally, we can expand the broadcasted dimensions to the requested shape
        semicontiguous = semicontiguous.expand(tensor.shape)
        return semicontiguous.view(shape)


def ravel_multi_index(multi_index: Sequence[torch.Tensor], dims: Sequence[int]) -> torch.Tensor:
    """
    Convert a multi-dimensional index into a flat index.

    Parameters
    ----------
    multi_index
        Sequence of integer index tensors.
    dims
        The shape of the tensor being indexed.

    Returns
    -------
    index
        Flattened index.
    """
    flat_index = multi_index[0]
    for idx, dim in zip(multi_index[1:], dims[1:], strict=True):
        flat_index = flat_index * dim + idx
    return flat_index


def broadcasted_rearrange(
    tensor: torch.Tensor,
    pattern: str,
    broadcasted_shape: Sequence[int] | None = None,
    *,
    reduce_views: bool = True,
    **axes_lengths: int,
) -> torch.Tensor:
    """Rearrange a tensor with broadcasting.

    Performs the einops rearrange or repeat operation on a tensor while preserving broadcasting.

    Rearranging is a smart element reordering for multidimensional tensors.
    This operation includes functionality of transpose (axes permutation),
    reshape (view), squeeze, unsqueeze, repeat, and tile functions.

    If a tensor has stride-0 dimensions, by default they will be preserved as stride-0
    if possible and not made contiguous, thus saving memory.
    If `reduce_views` is True, then stride-0 dimensions will be reduced to singleton dimensions after rearranging.
    Optionally performs broadcasting to a specified shape before rearranging.

    Examples
    --------
    ```python
    >>> tensor = torch.randn(1, 16, 1, 768, 256)
    >>> broadcasted_rearrange(tensor, '... (phase k1) k0 -> phase ... k1 k0', phase=8, reduce_views=False).shape
    torch.Size([8, 1, 16, 1, 96, 256])

    >>> tensor=torch.randn(1, 1, 1, 768, 1)
    >>> broadcasted_rearrange(tensor, '... (phase k1) k0 -> phase ... k1 k0',
    >>>    broadcasted_shape=(1, 16, 1, 768, 256), phase=8, reduce_views=False).shape
    torch.Size([8, 1, 16, 1, 96, 256]) # Behaves as-if the tensor was of shape (1, 16, 1, 768, 256)

    >>> tensor=torch.randn(1, 1, 1, 768, 1)
    >>> broadcasted_rearrange(tensor, '... (phase k1) k0 -> phase ... k1 k0',
    >>>    broadcasted_shape=(1, 16, 1, 768, 256) phase=8, reduce_views=True).shape
    torch.Size([8, 1, 1, 1, 96, 1]) # Dimensions that are stride-0 are reduced to singleton dimensions
    ```

    Parameters
    ----------
    tensor
        The input tensor to rearrange.
    pattern
        The rearrange pattern. See `einops` documentation for more information.
    broadcasted_shape
        The shape to broadcast the tensor to before rearranging. If `None`, no additional broadcasting is performed.
    reduce_views
        If `True`, reduce stride-0 dimensions to singleton dimensions after rearranging.
    axes_lengths
        The lengths of the axes in the pattern. See `einops` documentation for more information.


    """
    tensor = tensor.broadcast_to(broadcasted_shape) if broadcasted_shape is not None else tensor
    # the broadcast-preservation is done by patching the reshape method of the einops backend
    einops._backends.TorchBackend.reshape = lambda _, tensor, shape: reshape_broadcasted(tensor, *shape)  # type: ignore[method-assign]
    new_tensor = einops.repeat(tensor, pattern, **axes_lengths)  # allows both repeat and rearrange
    del einops._backends.TorchBackend.reshape  # resets to inherited method
    if reduce_views:
        new_tensor = reduce_view(new_tensor)
    return new_tensor


def expand_dim(tensor: torch.Tensor, dim: int, size: int) -> torch.Tensor:
    """Expand a tensor in one dimension.

    Parameters
    ----------
    tensor
        The tensor to expand.
    dim
        The dimension to expand.
    size
        The size to expand to.
    """
    new_shape = list(tensor.shape)
    new_shape[dim] = size
    return tensor.expand(new_shape)


def broadcasted_concatenate(tensors: Sequence[torch.Tensor], dim: int, reduce_views: bool = True) -> torch.Tensor:
    """Concatenate tensors while preserving broadcasting.

    Parameters
    ----------
    tensors
        The tensors to concatenate.
    dim
        The dimension to concatenate along.
    reduce_views
        If `True`, reduce stride-0 dimensions to singleton dimensions after concatenating.

    Returns
    -------
        The concatenated tensor.
    """
    n_dim = tensors[0].ndim
    if any(t.ndim != n_dim for t in tensors):
        raise ValueError('All tensors must have the same number of dimensions')
    if not (-n_dim <= dim < n_dim):
        raise ValueError(f'Dimension {dim} out of range for tensor of dimension {n_dim}')
    dim = dim % n_dim

    broadcasted_shape = []
    idx = []
    for n in range(n_dim):
        if n != dim and any(t.size(n) != tensors[0].size(n) for t in tensors):
            raise ValueError('All shapes must have the same size except for the concatenation dimension')
        if n == dim:
            idx.append(slice(None))  # keep all elements
            broadcasted_shape.append(-1)
        elif all(t.stride(n) == 0 for t in tensors):
            broadcasted_shape.append(tensors[0].size(n))
            idx.append(slice(1))  # reduce to singleton
        else:
            broadcasted_shape.append(tensors[0].size(n))
            idx.append(slice(None))

    tensors = [t[tuple(idx)] for t in tensors]
    result = torch.cat(tensors, dim=dim)

    if not reduce_views:  # dimensions are already reduced, we would undo this here.
        result = result.expand(broadcasted_shape)
    return result



"""Iterative SENSE Reconstruction."""

from __future__ import annotations

from collections.abc import Callable

from mrpro.algorithms.reconstruction.RegularizedIterativeSENSEReconstruction import (
    RegularizedIterativeSENSEReconstruction,
)
from mrpro.data.CsmData import CsmData
from mrpro.data.DcfData import DcfData
from mrpro.data.KData import KData
from mrpro.data.KNoise import KNoise
from mrpro.operators.LinearOperator import LinearOperator


class IterativeSENSEReconstruction(RegularizedIterativeSENSEReconstruction):
    r"""Iterative SENSE reconstruction.

    This algorithm solves the problem :math:`\min_{x} ||Ax - y||_2^2`
    by using a conjugate gradient algorithm to solve
    :math:`H x = b` with :math:`H = A^H A` and :math:`b = A^H y` where :math:`A` is the acquisition model
    (coil sensitivity maps, Fourier operator, k-space sampling), :math:`y` is the acquired k-space data [PRU2001]_ .

    Note: In [PRU2001]_ , a k-space weighting of the loss is used and a k-space filter is applied as a final step to
    null all k-space values outside the k-space coverage. This is not done here.

    .. [PRU2001] Pruessmann K, Weiger M, Boernert P, and Boesiger P (2001), Advances in sensitivity encoding with
       arbitrary k-space trajectories. MRI 46, 638-651. https://doi.org/10.1002/mrm.1241

    """

    n_iterations: int
    """Number of CG iterations."""

    def __init__(
        self,
        kdata: KData | None = None,
        fourier_op: LinearOperator | None = None,
        csm: Callable | CsmData | None = CsmData.from_idata_walsh,
        noise: KNoise | None = None,
        dcf: DcfData | None = None,
        *,
        n_iterations: int = 5,
    ) -> None:
        """Initialize IterativeSENSEReconstruction.

        For a regularized version of the iterative SENSE algorithm
        please see `~mrpro.algorithms.reconstruction.RegularizedIterativeSENSEReconstruction`.

        Parameters
        ----------
        kdata
            If `kdata` is provided and `fourier_op` or `dcf` are `None`, then `fourier_op` and `dcf` are estimated
            based on `kdata`. Otherwise `fourier_op` and `dcf` are used as provided.
        fourier_op
            Instance of the `~mrpro.operators.FourierOp` used for reconstruction.
            If `None`, set up based on `kdata`.
        csm
            Sensitivity maps for coil combination. If `None`, no coil combination is carried out, i.e. images for each
            coil are returned. If a `Callable` is provided, coil images are reconstructed using the adjoint of the
            `~mrpro.operators.FourierOp` (including density compensation) and then sensitivity maps are calculated
            using the `Callable`. For this, `kdata` needs also to be provided.
            For examples have a look at the `mrpro.data.CsmData` class e.g. `~mrpro.data.CsmData.from_idata_walsh`
            or `~mrpro.data.CsmData.from_idata_inati`.
        noise
            Noise used for prewhitening. If `None`, no prewhitening is performed
        dcf
            K-space sampling density compensation. Only used to obtain a good starting point for the CG algorithm as
            by using the scaled density compensated direct reconstruction [FESSLER2010]_.
        n_iterations
            Number of CG iterations

        Raises
        ------
        `ValueError`
            If the `kdata` and `fourier_op` are `None` or if `csm` is a `Callable` but `kdata` is None.
        """
        super().__init__(kdata, fourier_op, csm, noise, dcf, n_iterations=n_iterations, regularization_weight=0)



"""Fill tensor in-place along a specified dimension with increasing integers."""

import torch


def fill_range_(tensor: torch.Tensor, dim: int) -> None:
    """
    Fill tensor in-place along a specified dimension with increasing integers.

    Parameters
    ----------
    tensor
        The tensor to be modified in-place.

    dim
        The dimension along which to fill with increasing values.
    """
    if not -tensor.ndim <= dim < tensor.ndim:
        raise IndexError(f'Dimension {dim} is out of range for tensor with {tensor.ndim} dimensions.')

    dim = dim % tensor.ndim
    shape = [s if d == dim else 1 for d, s in enumerate(tensor.shape)]
    values = torch.arange(tensor.size(dim), device=tensor.device).reshape(shape)
    tensor[:] = values.expand_as(tensor)



from mrpro._version import __version__
from mrpro import algorithms, operators, data, phantoms, utils
__all__ = [
    "__version__",
    "algorithms",
    "data",
    "operators",
    "phantoms",
    "utils"
]



"""Some type hints that are used in multiple places in the codebase but not part of mrpro's public API."""

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeAlias

from typing_extensions import Any, Protocol, TypeVar, runtime_checkable

if TYPE_CHECKING:
    from dataclasses import Field
    from os import PathLike
    from types import EllipsisType
    from typing import ClassVar, TypeAlias

    import torch
    from numpy import ndarray
    from numpy._typing import _NestedSequence as NestedSequence
    from typing_extensions import ParamSpec, SupportsIndex, Unpack, overload

    # This matches the torch.Tensor indexer typehint
    _TorchIndexerTypeInner: TypeAlias = None | bool | int | slice | EllipsisType | torch.Tensor
    _SingleTorchIndexerType: TypeAlias = SupportsIndex | _TorchIndexerTypeInner | NestedSequence[_TorchIndexerTypeInner]
    TorchIndexerType: TypeAlias = tuple[_SingleTorchIndexerType, ...] | _SingleTorchIndexerType

    # This matches the numpy.ndarray indexer typehint
    _SingleNumpyIndexerType: TypeAlias = ndarray | SupportsIndex | None | slice | EllipsisType
    NumpyIndexerType: TypeAlias = tuple[_SingleNumpyIndexerType, ...] | _SingleNumpyIndexerType

    Tout = TypeVar('Tout', bound=tuple[torch.Tensor, ...], covariant=True)

    P = ParamSpec('P')
    Wrapped: TypeAlias = Callable[P, Tout]

    F = TypeVar('F', bound=Wrapped)

    class _EndomorphCallable(Protocol):
        """A callable with the same number of tensor inputs and outputs.

        This is a protocol for a callable that takes a variadic number of tensor inputs
        and returns the same number of tensor outputs.

        This is only implemented for up to 10 inputs, if more inputs are given, the return
        will be a variadic number of tensors.

        This Protocol is used to decorate functions with the `endomorph` decorator.
        """

        @overload
        def __call__(self, /, **kwargs) -> tuple[()]: ...
        @overload
        def __call__(self, x1: torch.Tensor, /, **kwargs) -> tuple[torch.Tensor]: ...

        @overload
        def __call__(self, x1: torch.Tensor, x2: torch.Tensor, /, **kwargs) -> tuple[torch.Tensor, torch.Tensor]: ...

        @overload
        def __call__(
            self, x1: torch.Tensor, x2: torch.Tensor, x3: torch.Tensor, /, **kwargs
        ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]: ...

        @overload
        def __call__(
            self, x1: torch.Tensor, x2: torch.Tensor, x3: torch.Tensor, x4: torch.Tensor, /, **kwargs
        ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: ...

        @overload
        def __call__(
            self, x1: torch.Tensor, x2: torch.Tensor, x3: torch.Tensor, x4: torch.Tensor, x5: torch.Tensor, /, **kwargs
        ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            /,
            **kwargs,
        ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            x7: torch.Tensor,
            /,
            **kwargs,
        ) -> tuple[
            torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor
        ]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            x7: torch.Tensor,
            x8: torch.Tensor,
            /,
            **kwargs,
        ) -> tuple[
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
        ]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            x7: torch.Tensor,
            x8: torch.Tensor,
            x9: torch.Tensor,
            /,
            **kwargs,
        ) -> tuple[
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
        ]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            x7: torch.Tensor,
            x8: torch.Tensor,
            x9: torch.Tensor,
            x10: torch.Tensor,
            /,
            **kwargs,
        ) -> tuple[
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
        ]: ...

        @overload
        def __call__(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            x3: torch.Tensor,
            x4: torch.Tensor,
            x5: torch.Tensor,
            x6: torch.Tensor,
            x7: torch.Tensor,
            x8: torch.Tensor,
            x9: torch.Tensor,
            x10: torch.Tensor,
            /,
            *args: torch.Tensor,
            **kwargs,
        ) -> tuple[
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            torch.Tensor,
            Unpack[tuple[torch.Tensor, ...]],
        ]: ...

        @overload
        def __call__(self, /, *args: torch.Tensor, **kwargs) -> tuple[torch.Tensor, ...]: ...

        def __call__(self, /, *args: torch.Tensor, **kwargs) -> tuple[torch.Tensor, ...]:
            """Apply the Function."""

    def endomorph(f: F, /) -> _EndomorphCallable:
        """Decorate a function to make it an endomorph callable.

        This adds overloads for N->N-Tensor signatures, for N<10.
        For >10 inputs, the return type will a tuple with >10 tensors.
        """
        return f

    class ReadSeekable(Protocol):
        def read(self, size: int | None = -1) -> bytes:
            """Read up to size bytes. Returns fewer if EOF is reached."""
            ...

        def seek(self, offset: int, whence: int = 0) -> int:
            """Change stream position and return the new absolute position."""
            ...

    FileOrPath: TypeAlias = ReadSeekable | str | PathLike

    @runtime_checkable
    class DataclassInstance(Protocol):
        """An instance of a dataclass."""

        __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


else:
    TorchIndexerType: TypeAlias = Any
    """Torch indexer type."""

    class NestedSequence(Protocol[TypeVar('T')]):
        """A nested sequence type."""

        ...

    @runtime_checkable
    class DataclassInstance(Protocol):
        """An instance of a dataclass."""

        __dataclass_fields__: Any

    NumpyIndexerType: TypeAlias = Any
    """Numpy indexer type."""

    def endomorph(f: Callable) -> Callable:
        """Decorate a function to make it an endomorph callable."""
        return f

    FileOrPath: TypeAlias = Any


__all__ = ['DataclassInstance', 'FileOrPath', 'NestedSequence', 'NumpyIndexerType', 'TorchIndexerType', 'endomorph']



"""MR noise measurements class."""

from collections.abc import Callable
from pathlib import Path

import ismrmrd
import torch
from einops import repeat
from typing_extensions import Self

from mrpro.data.acq_filters import is_noise_acquisition
from mrpro.data.Dataclass import Dataclass


class KNoise(Dataclass):
    """MR raw data / k-space data class for noise measurements."""

    data: torch.Tensor
    """K-space data of noise measurements. Shape `(*other coils k2 k1 k0)`"""

    @classmethod
    def from_file(
        cls, filename: str | Path, dataset_idx: int = -1, acquisition_filter_criterion: Callable = is_noise_acquisition
    ) -> Self:
        """Load noise measurements from ISMRMRD file.

        Parameters
        ----------
        filename
            Path to the ISMRMRD file
        dataset_idx
            Index of the dataset to load (converter creates dataset, dataset_1, ...)
        acquisition_filter_criterion
            function which returns True if an acquisition should be included in KNoise
        """
        # Can raise FileNotFoundError
        with ismrmrd.File(filename, 'r') as file:
            ds = file[list(file.keys())[dataset_idx]]
            acquisitions = ds.acquisitions[:]

        # Read out noise measurements
        acquisitions = [acq for acq in acquisitions if acquisition_filter_criterion(acq)]
        if len(acquisitions) == 0:
            raise ValueError(f'No noise measurements found in {filename}')
        noise_data = torch.stack([torch.as_tensor(acq.data, dtype=torch.complex64) for acq in acquisitions])

        # Reshape to standard dimensions
        noise_data = repeat(noise_data, '... coils k0->... coils k2 k1 k0', k1=1, k2=1)

        return cls(noise_data)



"""Linear Operators."""

from __future__ import annotations

import functools
import operator
from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import cast, no_type_check

import torch
from typing_extensions import Any, Unpack, overload

import mrpro.operators
from mrpro.operators.Operator import Operator, OperatorComposition, OperatorSum, Tin2


class _AutogradWrapper(torch.autograd.Function):
    """Wrap forward and adjoint functions for autograd."""

    # If the forward and adjoint implementation are vmap-compatible,
    # the function can be marked as such to enable vmap support.
    generate_vmap_rule = True

    @staticmethod
    def forward(
        fw: Callable[[torch.Tensor], torch.Tensor],
        bw: Callable[[torch.Tensor], torch.Tensor],  # noqa: ARG004
        x: torch.Tensor,
    ) -> torch.Tensor:
        return fw(x)

    @staticmethod
    def setup_context(
        ctx: Any,  # noqa: ANN401
        inputs: tuple[Callable[[torch.Tensor], torch.Tensor], Callable[[torch.Tensor], torch.Tensor], torch.Tensor],
        _output: torch.Tensor,
    ) -> None:
        ctx.fw, ctx.bw, _ = inputs
        return None

    @staticmethod
    def backward(ctx: Any, *grad_output: torch.Tensor) -> tuple[None, None, torch.Tensor]:  # noqa: ANN401
        return None, None, _AutogradWrapper.apply(ctx.bw, ctx.fw, grad_output[0])

    @staticmethod
    def jvp(ctx: Any, *grad_inputs: Any) -> torch.Tensor:  # noqa: ANN401
        return _AutogradWrapper.apply(ctx.fw, ctx.bw, grad_inputs[-1])


class LinearOperator(Operator[torch.Tensor, tuple[torch.Tensor]]):
    """General Linear Operator.

    LinearOperators have exactly one input tensors and one output tensor,
    and fulfill :math:`f(ax + by) = af(x) + bf(y)`
    with :math:`a`, :math:`b` scalars and :math:`x`, :math:`y` tensors.

    LinearOperators can be composed, added, multiplied, applied to tensors.
    LinearOperators have an `~LinearOperator.H` property that returns the adjoint operator,
    and a `~LinearOperator.gram` property that returns the Gram operator.

    Subclasses must implement the forward and adjoint methods.
    When subclassing, the `adjoint_as_backward` class attribute can be set to `True`::

            class MyOperator(LinearOperator, adjoint_as_backward=True):
                ...

    This will make pytorch use the adjoint method as the backward method of the forward,
    and the forward method as the backward method of the adjoint, avoiding the need to
    have differentiable forward and adjoint methods.
    """

    @no_type_check
    def __init_subclass__(cls, adjoint_as_backward: bool = False, **kwargs: Any) -> None:  # noqa: ANN401
        """Wrap the forward and adjoint functions for autograd.

        This will  wrap the forward and adjoint functions for autograd,
        and use the adjoint function as the backward function of the forward and vice versa.

        Parameters
        ----------
        adjoint_as_backward
            if True, the adjoint function is used as the backward function,
            else automatic differentiation of the forward function is used.
        kwargs
            additional keyword arguments, passed to the super class
        """
        if adjoint_as_backward and not hasattr(cls, '_saved_forward'):
            cls._saved_forward, cls._saved_adjoint = cls.forward, cls.adjoint
            cls.forward = functools.update_wrapper(
                lambda self, x: (
                    _AutogradWrapper.apply(lambda x: self._saved_forward(x)[0], lambda x: self._saved_adjoint(x)[0], x),
                ),
                cls.forward,
            )
            cls.adjoint = functools.update_wrapper(
                lambda self, x: (
                    _AutogradWrapper.apply(lambda x: self._saved_adjoint(x)[0], lambda x: self._saved_forward(x)[0], x),
                ),
                cls.adjoint,
            )
        super().__init_subclass__(**kwargs)

    @abstractmethod
    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint of the operator."""
        ...

    @property
    def H(self) -> LinearOperator:  # noqa: N802
        """Adjoint operator.

        Obtains the adjoint of an instance of this operator as an `AdjointLinearOperator`,
        which itself is a an `LinearOperator` that can be applied to tensors.

        Note: ``linear_operator.H.H == linear_operator``
        """
        return AdjointLinearOperator(self)

    def operator_norm(
        self,
        initial_value: torch.Tensor,
        dim: Sequence[int] | None,
        max_iterations: int = 20,
        relative_tolerance: float = 1e-4,
        absolute_tolerance: float = 1e-5,
        callback: Callable[[torch.Tensor], None] | None = None,
    ) -> torch.Tensor:
        """Power iteration for computing the operator norm of the operator.

        Parameters
        ----------
        initial_value
            initial value to start the iteration; must be element of the domain.
            if the initial value contains a zero-vector for one of the considered problems,
            the function throws an `ValueError`.
        dim
            The dimensions of the tensors on which the operator operates. The choice of `dim` determines how
            the operator norm is inperpreted. For example, for a matrix-vector multiplication with a batched matrix
            tensor of shape `(batch1, batch2, row, column)` and a batched input tensor of shape `(batch1, batch2, row)`:

            * If `dim=None`, the operator is considered as a block diagonal matrix with batch1*batch2 blocks
              and the result is a tensor containing a single norm value (shape `(1, 1, 1)`).

            * If `dim=(-1)`, `batch1*batch2` matrices are considered, and for each a separate operator norm is computed.

            * If `dim=(-2,-1)`, `batch1` matrices with `batch2` blocks are considered, and for each matrix a
              separate operator norm is computed.

            Thus, the choice of `dim` determines implicitly determines the domain of the operator.
        max_iterations
            maximum number of iterations
        relative_tolerance
            absolute tolerance for the change of the operator-norm at each iteration;
            if set to zero, the maximal number of iterations is the only stopping criterion used to stop
            the power iteration.
        absolute_tolerance
            absolute tolerance for the change of the operator-norm at each iteration;
            if set to zero, the maximal number of iterations is the only stopping criterion used to stop
            the power iteration.
        callback
            user-provided function to be called at each iteration

        Returns
        -------
            An estimaton of the operator norm. Shape corresponds to the shape of the input tensor `initial_value`
            with the dimensions specified in `dim` reduced to a single value.
            The pointwise multiplication of `initial_value` with the result of the operator norm will always
            be well-defined.
        """
        if max_iterations < 1:
            raise ValueError('The number of iterations should be larger than zero.')

        dim = tuple(dim) if dim is not None else dim  # must be tuple or None for torch.sum

        norm_initial_value = torch.linalg.vector_norm(initial_value, dim=dim, keepdim=True)
        if not (norm_initial_value > 0).all():
            if dim is None:
                raise ValueError('The initial value for the iteration should be different from the zero-vector.')
            else:
                raise ValueError(
                    'Found at least one zero-vector as starting point. \
                    For each of the considered operators, the initial value for the iteration \
                    should be different from the zero-vector.'
                )

        # creaty dummy operator norm value that cannot be correct because by definition, the
        # operator norm is a strictly positive number. This ensures that the first time the
        # change between the old and the new estimate of the operator norm is non-zero and
        # thus prevents the loop from exiting despite a non-correct estimate.
        op_norm_old = torch.zeros(*tuple([1 for _ in range(initial_value.ndim)]), device=initial_value.device)

        gram = self.gram  # self.H@self

        vector = initial_value
        for _ in range(max_iterations):
            # apply the operator to the vector
            (vector_new,) = gram(vector)

            # compute estimate of the operator norm
            product = vector.real * vector_new.real
            if vector.is_complex() and vector_new.is_complex():
                product = product + vector.imag * vector_new.imag
            op_norm = product.sum(dim, keepdim=True).sqrt()

            # check if stopping criterion is fulfillfed; if not continue the iteration
            if (absolute_tolerance > 0 or relative_tolerance > 0) and torch.isclose(
                op_norm, op_norm_old, atol=absolute_tolerance, rtol=relative_tolerance
            ).all():
                break

            # normalize vector
            vector = vector_new / torch.linalg.vector_norm(vector_new, dim=dim, keepdim=True)
            op_norm_old = op_norm

            if callback is not None:
                callback(op_norm)

        return op_norm

    @overload
    def __matmul__(self, other: LinearOperator) -> LinearOperator: ...

    @overload
    def __matmul__(
        self, other: Operator[Unpack[Tin2], tuple[torch.Tensor,]] | Operator[Unpack[Tin2], tuple[torch.Tensor, ...]]
    ) -> Operator[Unpack[Tin2], tuple[torch.Tensor,]]: ...

    def __matmul__(
        self,
        other: Operator[Unpack[Tin2], tuple[torch.Tensor,]]
        | LinearOperator
        | Operator[Unpack[Tin2], tuple[torch.Tensor, ...]],
    ) -> (
        Operator[Unpack[Tin2], tuple[torch.Tensor,]] | LinearOperator | Operator[Unpack[Tin2], tuple[torch.Tensor, ...]]
    ):
        """Operator composition.

        Returns ``lambda x: self(other(x))``
        """
        if isinstance(other, mrpro.operators.IdentityOp):
            # neutral element of composition
            return self
        elif isinstance(self, mrpro.operators.IdentityOp):
            return other
        elif isinstance(self, mrpro.operators.ZeroOp) or isinstance(other, mrpro.operators.ZeroOp):
            # zero operator composition with any operator is zero
            return mrpro.operators.ZeroOp()
        elif isinstance(other, LinearOperator):
            # LinearOperator@LinearOperator is linear
            return LinearOperatorComposition(self, other)
        elif isinstance(other, Operator):
            # cast due to https://github.com/python/mypy/issues/16335
            return OperatorComposition(self, cast(Operator[Unpack[Tin2], tuple[torch.Tensor,]], other))
        return NotImplemented

    def __radd__(self, other: torch.Tensor | complex) -> LinearOperator:
        """Operator addition.

        Returns ``lambda x: self(x) + other*x``
        """
        return self + other

    @overload  # type: ignore[override]
    def __add__(self, other: LinearOperator | torch.Tensor | complex) -> LinearOperator: ...

    @overload
    def __add__(
        self, other: Operator[torch.Tensor, tuple[torch.Tensor,]]
    ) -> Operator[torch.Tensor, tuple[torch.Tensor,]]: ...

    def __add__(
        self, other: Operator[torch.Tensor, tuple[torch.Tensor,]] | LinearOperator | torch.Tensor | complex
    ) -> Operator[torch.Tensor, tuple[torch.Tensor,]] | LinearOperator:
        """Operator addition.

        Returns ``lambda x: self(x) + other(x)`` if other is a operator,
        ``lambda x: self(x) + other`` if other is a tensor
        """
        if isinstance(other, torch.Tensor | complex | int | float):
            # tensor addition
            return LinearOperatorSum(self, mrpro.operators.IdentityOp() * other)
        elif isinstance(self, mrpro.operators.ZeroOp):
            # neutral element of addition
            return other
        elif isinstance(other, mrpro.operators.ZeroOp):
            # neutral element of addition
            return self
        elif isinstance(other, LinearOperator):
            # sum of LinearOperators is linear
            return LinearOperatorSum(self, other)
        elif isinstance(other, Operator):
            # for general operators
            return OperatorSum(self, other)
        else:
            return NotImplemented

    def __mul__(self, other: torch.Tensor | complex) -> LinearOperator:
        """Operator elementwise left multiplication with tensor/scalar.

        Returns ``lambda x: self(x*other)``
        """
        if isinstance(other, complex | float | int):
            if other == 0:
                return mrpro.operators.ZeroOp()
            if other == 1:
                return self
            else:
                return LinearOperatorElementwiseProductLeft(self, other)
        elif isinstance(other, torch.Tensor):
            return LinearOperatorElementwiseProductLeft(self, other)
        else:
            return NotImplemented

    def __rmul__(self, other: torch.Tensor | complex) -> LinearOperator:
        """Operator elementwise right multiplication with tensor/scalar.

        Returns ``lambda x: other*self(x)``
        """
        if isinstance(other, complex | float | int):
            if other == 0:
                return mrpro.operators.ZeroOp()
            if other == 1:
                return self
            else:
                return LinearOperatorElementwiseProductRight(self, other)
        elif isinstance(other, torch.Tensor):
            return LinearOperatorElementwiseProductRight(self, other)
        else:
            return NotImplemented

    def __or__(self, other: LinearOperator) -> mrpro.operators.LinearOperatorMatrix:
        """Horizontal stacking of two LinearOperators.

        ``A|B`` is a `~mrpro.operators.LinearOperatorMatrix` with two columns,
        with ``(A|B)(x1,x2) == A(x1) + B(x2)``.
        See `mrpro.operators.LinearOperatorMatrix` for more information.
        """
        if not isinstance(other, LinearOperator):
            return NotImplemented
        operators = [[self, other]]
        return mrpro.operators.LinearOperatorMatrix(operators)

    def __mod__(self, other: LinearOperator) -> mrpro.operators.LinearOperatorMatrix:
        """Vertical stacking of two LinearOperators.

        ``A%B`` is a `~mrpro.operators.LinearOperatorMatrix` with two rows,
        with ``(A%B)(x) == (A(x), B(x))``.
        See `mrpro.operators.LinearOperatorMatrix` for more information.
        """
        if not isinstance(other, LinearOperator):
            return NotImplemented
        operators = [[self], [other]]
        return mrpro.operators.LinearOperatorMatrix(operators)

    @property
    def gram(self) -> LinearOperator:
        """Gram operator.

        For a LinearOperator :math:`A`, the self-adjoint Gram operator is defined as :math:`A^H A`.

        .. note::
           This is the inherited default implementation.
        """
        return self.H @ self


class LinearOperatorComposition(LinearOperator):
    """Linear operator composition.

    Performs operator1(operator2(x))
    """

    def __init__(self, operator1: LinearOperator, operator2: LinearOperator) -> None:
        """Linear operator composition initialization.

        Returns ``lambda x: operator1(operator2(x))``

        Parameters
        ----------
        operator1
            outer operator
        operator2
            inner operator
        """
        super().__init__()
        self._operator1 = operator1
        self._operator2 = operator2

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Linear operator composition.

        Performs operator1(operator2(x)).

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the composed operation.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of LinearOperatorComposition.

        .. note::
            Prefer calling the instance of the LinearOperatorComposition operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return self._operator1(*self._operator2(x))

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint of the linear operator composition.

        The adjoint of (operator1 @ operator2) is (operator2.H @ operator1.H).

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the adjoint operation.
        """
        # (AB)^H = B^H A^H
        return self._operator2.adjoint(*self._operator1.adjoint(x))

    @property
    def gram(self) -> LinearOperator:
        """Gram operator."""
        # (AB)^H(AB) = B^H (A^H A) B
        return self._operator2.H @ self._operator1.gram @ self._operator2


class LinearOperatorSum(LinearOperator):
    """Linear operator addition."""

    _operators: list[LinearOperator]

    def __init__(self, operator1: LinearOperator, /, *other_operators: LinearOperator):
        """Linear operator addition initialization."""
        super().__init__()
        ops: list[LinearOperator] = []
        for op in (operator1, *other_operators):
            if isinstance(op, LinearOperatorSum):
                ops.extend(op._operators)
            else:
                ops.append(op)
        self._operators = cast(list[LinearOperator], torch.nn.ModuleList(ops))

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Linear operator addition.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of applying the sum of operators.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of LinearOperatorSum.

        .. note::
            Prefer calling the instance of the LinearOperatorSum operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (functools.reduce(operator.add, (op(x)[0] for op in self._operators)),)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint of the linear operator addition.

        The adjoint of (A+B) is (A.H + B.H).

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of applying the sum of adjoint operators.
        """
        # (A+B)^H = A^H + B^H
        return (functools.reduce(operator.add, (op.adjoint(x)[0] for op in self._operators)),)


class LinearOperatorElementwiseProductRight(LinearOperator):
    """Linear operator elementwise right multiplication with a tensor.

    Performs Tensor*LinearOperator(x)
    """

    def __init__(self, operator: LinearOperator, scalar: torch.Tensor | complex):
        """Linear operator elementwise right multiplication initialization."""
        super().__init__()
        self._operator = operator
        self._scalar = scalar

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Linear operator elementwise right multiplication.

        Performs ``scalar * Operator(x)``.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the elementwise multiplication.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of LinearOperatorElementwiseProductRight.

        .. note::
            Prefer calling the instance of the LinearOperatorElementwiseProductRight operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        (out,) = self._operator(x)
        return (out * self._scalar,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint of elementwise right multiplication.

        The adjoint of (scalar * A) is (scalar_conj * A.H).

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the adjoint operation.
        """
        conj = self._scalar.conj() if isinstance(self._scalar, torch.Tensor) else self._scalar.conjugate()
        return self._operator.adjoint(x * conj)

    @property
    def gram(self) -> LinearOperator:
        """Gram Operator."""
        if isinstance(self._scalar, torch.Tensor):
            factor: torch.Tensor | complex = self._scalar.conj() * self._scalar
            if self._scalar.numel() > 1:
                # only scalars can be moved outside the linear operator
                return self._operator.H @ (factor * self._operator)
        else:
            factor = self._scalar.conjugate() * self._scalar
        return factor * self._operator.gram


class LinearOperatorElementwiseProductLeft(LinearOperator):
    """Operator elementwise left multiplication with a tensor.

    Performs LinearOperator(Tensor*x)
    """

    def __init__(self, operator: LinearOperator, scalar: torch.Tensor | complex) -> None:
        """Linear operator elementwise left multiplication initialization."""
        super().__init__()
        self._operator = operator
        self._scalar = scalar

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Operator elementwise left multiplication with a tensor.

        Performs ``Operator(scalar * x)``.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the operator acting on the scaled tensor.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of LinearOperatorElementwiseProductLeft.

        .. note::
            Prefer calling the instance of the LinearOperatorElementwiseProductLeft operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return self._operator(x * self._scalar)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Adjoint of elementwise left multiplication.

        The adjoint of (A * scalar) is (A.H * scalar_conj).

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Result of the adjoint operation.
        """
        conj = self._scalar.conj() if isinstance(self._scalar, torch.Tensor) else self._scalar.conjugate()
        return (self._operator.adjoint(x)[0] * conj,)

    @property
    def gram(self) -> LinearOperator:
        """Gram Operator."""
        conj = self._scalar.conj() if isinstance(self._scalar, torch.Tensor) else self._scalar.conjugate()
        return conj * self._operator.gram * self._scalar


class AdjointLinearOperator(LinearOperator):
    """Adjoint of a LinearOperator."""

    def __init__(self, operator: LinearOperator) -> None:
        """Initialize the adjoint of a LinearOperator."""
        super().__init__()
        self._operator = operator

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the original LinearOperator.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Tensor resulting from the adjoint operation.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of AdjointLinearOperator.

        .. note::
            Prefer calling the instance of the AdjointLinearOperator operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return self._operator.adjoint(x)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the adjoint, i.e. the original LinearOperator.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Tensor resulting from applying the original operator.
        """
        return self._operator.forward(x)

    @property
    def H(self) -> LinearOperator:  # noqa: N802
        """Adjoint of adjoint operator, i.e. original LinearOperator."""
        return self._operator



"""Class for Finite Difference Operator."""

from collections.abc import Sequence
from typing import Literal

import torch

from mrpro.operators.LinearOperator import LinearOperator
from mrpro.utils.filters import filter_separable


class FiniteDifferenceOp(LinearOperator):
    """Finite Difference Operator."""

    @staticmethod
    def finite_difference_kernel(mode: Literal['central', 'forward', 'backward']) -> torch.Tensor:
        """Finite difference kernel.

        Parameters
        ----------
        mode
            String specifying kernel type

        Returns
        -------
            Finite difference kernel

        Raises
        ------
        `ValueError`
            If mode is not central, forward, backward or doublecentral
        """
        if mode == 'central':
            kernel = torch.tensor((-1, 0, 1)) / 2
        elif mode == 'forward':
            kernel = torch.tensor((0, -1, 1))
        elif mode == 'backward':
            kernel = torch.tensor((-1, 1, 0))
        else:
            raise ValueError(f'mode should be one of (central, forward, backward), not {mode}')
        return kernel

    def __init__(
        self,
        dim: Sequence[int],
        mode: Literal['central', 'forward', 'backward'] = 'central',
        pad_mode: Literal['zeros', 'circular'] = 'zeros',
    ) -> None:
        """Finite difference operator.

        Parameters
        ----------
        dim
            Dimension along which finite differences are calculated.
        mode
            Type of finite difference operator
        pad_mode
            Padding to ensure output has the same size as the input
        """
        super().__init__()
        self.dim = dim
        self.pad_mode: Literal['constant', 'circular'] = 'constant' if pad_mode == 'zeros' else pad_mode
        self.kernel = self.finite_difference_kernel(mode)

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward finite difference operation.

        Calculates finite differences of the input tensor `x` along the dimensions
        specified during initialization. The results for each dimension are stacked
        along the first dimension of the output tensor.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Tensor containing the finite differences of `x` along the specified
            dimensions, stacked along the first dimension.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of FiniteDifferenceOp.

        .. note::
            Prefer calling the instance of the FiniteDifferenceOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (
            torch.stack(
                [
                    filter_separable(x, (self.kernel,), dim=(dim,), pad_mode=self.pad_mode, pad_value=0.0)
                    for dim in self.dim
                ]
            ),
        )

    def adjoint(self, y: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply adjoint finite difference operation.

        This operation is the adjoint of the forward finite difference calculation.
        It takes a tensor `y` (which is assumed to be the output of the forward pass,
        i.e., finite differences stacked along the first dimension) and computes
        the sum of the adjoints of the individual directional finite difference operations.

        Parameters
        ----------
        y
            Input tensor, representing finite differences stacked along the first dimension.
            The size of the first dimension must match the number of dimensions
            specified for the operator.

        Returns
        -------
            Result of the adjoint finite difference operation.

        Raises
        ------
        ValueError
            If the first dimension of `y` does not match the number of
            dimensions along which finite differences were calculated.
        """
        if y.shape[0] != len(self.dim):
            raise ValueError('Fist dimension of input tensor has to match the number of finite difference directions.')
        return (
            torch.sum(
                torch.stack(
                    [
                        filter_separable(
                            yi,
                            (torch.flip(self.kernel, dims=(-1,)),),
                            dim=(dim,),
                            pad_mode=self.pad_mode,
                            pad_value=0.0,
                        )
                        for dim, yi in zip(self.dim, y, strict=False)
                    ]
                ),
                dim=0,
            ),
        )



"""Interpolation of data tensor."""

from collections.abc import Callable, Sequence
from typing import Literal

import torch


def interpolate(
    x: torch.Tensor, size: Sequence[int], dim: Sequence[int], mode: Literal['nearest', 'linear'] = 'linear'
) -> torch.Tensor:
    """Interpolate the tensor x along the axes dim to the new size.

    Parameters
    ----------
    x
        Tensor to interpolate
    size
        New size of the tensor
    dim
        Axes to interpolate over. Must have the same length as size.
    mode
        Interpolation mode.

    Returns
    -------
        The interpolated tensor, with the new size.

    """
    if len(dim) != len(size):
        raise ValueError('Must provide matching length size and dim arguments.')

    # normalize dim to allow negative indexing in input
    dim = tuple([a % x.ndim for a in dim])
    if len(dim) != len(set(dim)):
        raise ValueError(f'Dim must be unique. Normalized dims are {dim}')

    # return input tensor if old and new size match
    if all(x.shape[d] == s for s, d in zip(size, dim, strict=True)):
        return x

    # torch.nn.functional.interpolate only available for real tensors
    # moveaxis is not implemented for batched tensors, so vmap would fail, thus we use permute.
    x_real = torch.view_as_real(x).permute(-1, *range(x.ndim)) if x.is_complex() else x
    dim = [d + 1 for d in dim] if x.is_complex() else dim

    for s, d in zip(size, dim, strict=True):
        if s != x_real.shape[d]:
            idx = list(range(x_real.ndim))
            # swapping the last axis and the axis to filter over
            idx[d], idx[-1] = idx[-1], idx[d]
            x_real = x_real.permute(idx)
            x_real = torch.nn.functional.interpolate(x_real.flatten(end_dim=-3), size=s, mode=mode).reshape(
                *x_real.shape[:-1], -1
            )
            # for a single permutation, this undoes the permutation
            x_real = x_real.permute(idx)
    return torch.view_as_complex(x_real.permute(*range(1, x.ndim + 1), 0).contiguous()) if x.is_complex() else x_real


def apply_lowres(function: Callable[[torch.Tensor], torch.Tensor], size: Sequence[int], dim: Sequence[int]) -> Callable:
    """Apply function f on low-res version of tensor x and then return the upsampled f(x).

    Parameters
    ----------
    function
        Function to be applied on low-resolution version of tensor.
    size
        Low-resolution size of tensor.
    dim
        Low-resolution axes. Must have the same length as size.

    Returns
    -------
        Function which downsamples tensor, applies function and upsamples result.
    """

    def apply_to_lowres_data(x: torch.Tensor) -> torch.Tensor:
        """Downsample tensor, apply function and upsample result.

        Parameters
        ----------
        x
            Input tensor.

        Returns
        -------
            Tensor in original size with function applied.
        """
        x_lowres = interpolate(x, size, dim)
        x_lowres = function(x_lowres)
        return interpolate(x_lowres, [x.shape[d] for d in dim], dim)

    return apply_to_lowres_data



"""Signal Model Operators."""

import torch
from typing_extensions import TypeVarTuple, Unpack

from mrpro.operators.Operator import Operator

Tin = TypeVarTuple('Tin')


# SignalModel has multiple inputs and one output
class SignalModel(Operator[Unpack[Tin], tuple[torch.Tensor,]]):
    """Signal Model Operator."""



"""PCA Compression Operator."""

import einops
import torch
from einops import repeat

from mrpro.operators.LinearOperator import LinearOperator


class PCACompressionOp(LinearOperator):
    """PCA based compression operator."""

    def __init__(
        self,
        data: torch.Tensor,
        n_components: int,
    ) -> None:
        """Construct a PCA based compression operator.

        The operator carries out an SVD followed by a threshold of the `n_components` largest values along the last
        dimension of a data with shape `(*other, joint_dim, compression_dim)`.
        A single SVD is carried out for everything along `joint_dim`. `other` are batch dimensions.

        Consider combining this operator with `~mrpro.operators.RearrangeOp` to make sure the data is
        in the correct shape before applying.

        Parameters
        ----------
        data
            Data of shape `(*other, joint_dim, compression_dim)` to be used to find the principal components.
        n_components
            Number of principal components to keep along the compression_dim.
        """
        super().__init__()
        # different compression matrices along the *other dimensions
        data = data - data.mean(-1, keepdim=True)
        correlation = einops.einsum(data, data.conj(), '... joint comp1, ... joint comp2 -> ... comp1 comp2')
        _, _, v = torch.svd(correlation)
        # add joint_dim along which the the compression is the same
        v = repeat(v, '... comp1 comp2 -> ... joint_dim comp1 comp2', joint_dim=1)
        self._compression_matrix = v[..., :n_components, :].clone()

    def __call__(self, data: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply PCA-based compression to the input data.

        The data is projected onto the principal components determined during
        the operator's initialization.

        Parameters
        ----------
        data
            Input data to be compressed. Expected shape is
            `(*other, joint_dim, compression_dim)`.

        Returns
        -------
            Compressed data, with shape `(*other, joint_dim, n_components)`.
        """
        return super().__call__(data)

    def forward(self, data: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of PCACompressionOp.

        .. note::
            Prefer calling the instance of the PCACompressionOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        try:
            result = (self._compression_matrix @ data.unsqueeze(-1)).squeeze(-1)
        except RuntimeError as e:
            raise RuntimeError(
                'Shape mismatch in adjoint Compression: '
                f'Matrix {tuple(self._compression_matrix.shape)} '
                f'cannot be multiplied with Data {tuple(data.shape)}.'
            ) from e
        return (result,)

    def adjoint(self, data: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of PCA-based compression (expansion).

        The data, assumed to be in the compressed principal component space,
        is projected back to the original data space using the hermitian
        transpose of the compression matrix.

        Parameters
        ----------
        data
            Compressed input data. Expected shape is
            `(*other, joint_dim, n_components)`.

        Returns
        -------
            Expanded data, with shape `(*other, joint_dim, compression_dim)`.
        """
        try:
            result = (self._compression_matrix.mH @ data.unsqueeze(-1)).squeeze(-1)
        except RuntimeError as e:
            raise RuntimeError(
                'Shape mismatch in adjoint Compression: '
                f'Matrix^H {tuple(self._compression_matrix.mH.shape)} '
                f'cannot be multiplied with Data {tuple(data.shape)}.'
            ) from e
        return (result,)



"""ADAM for solving non-linear minimization problems."""

from collections.abc import Callable, Sequence

import torch
from torch.optim import Adam, AdamW

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.operators.Operator import OperatorType


def adam(
    f: OperatorType,
    initial_parameters: Sequence[torch.Tensor],
    *,
    max_iterations: int = 100,
    learning_rate: float = 1e-3,
    tolerance_change: float = 0,
    betas: tuple[float, float] = (0.9, 0.999),
    eps: float = 1e-8,
    weight_decay: float = 0,
    amsgrad: bool = False,
    decoupled_weight_decay: bool = False,
    callback: Callable[[OptimizerStatus], bool | None] | None = None,
) -> tuple[torch.Tensor, ...]:
    r"""Adam for non-linear minimization problems.

    Adam [KING2015]_ (Adaptive Moment Estimation) is a first-order optimization algorithm that adapts learning rates
    for each parameter using estimates of the first and second moments of the gradients.

    The parameter update rule is:

    .. math::

        m_t &= \beta_1 m_{t-1} + (1 - \beta_1) g_t \\
        v_t &= \beta_2 v_{t-1} + (1 - \beta_2) g_t^2 \\
        \hat{m}_t &= \frac{m_t}{1 - \beta_1^t}, \quad \hat{v}_t = \frac{v_t}{1 - \beta_2^t} \\
        \theta_{t+1} &= \theta_t - \frac{\eta}{\sqrt{\hat{v}_t} + \epsilon} \hat{m}_t

    where
    :math:`g_t` is the gradient at step :math:`t`,
    :math:`m_t` and :math:`v_t` are biased estimates of the first and second moments,
    :math:`\hat{m}_t` and :math:`\hat{v}_t` are bias-corrected estimates,
    :math:`\eta` is the learning rate,
    :math:`\epsilon` is a small constant for numerical stability,
    :math:`\beta_1` and :math:`\beta_2` are decay rates for the moment estimates.

    Steps of the Adam algorithm:

    1. Initialize parameters and moment estimates (:math:`m_0`, :math:`v_0`).
    2. Compute the gradient of the objective function.
    3. Compute bias-corrected estimates of the moments :math:`\hat{m}_t` and :math:`\hat{v}_t`.
    4. Update parameters using the adaptive step size.

    This function wraps PyTorch's :class:`torch.optim.Adam` and :class:`torch.optim.AdamW` implementations,
    supporting both standard Adam and decoupled weight decay regularization (AdamW) [LOS2019]_

    References
    ----------
    .. [KING2015] Kingma DP, Ba J (2015) Adam: A Method for Stochastic Optimization. ICLR.
       https://doi.org/10.48550/arXiv.1412.6980
    .. [LOS2019] Loshchilov I, Hutter F (2019) Decoupled Weight Decay Regularization. ICLR.
       https://doi.org/10.48550/arXiv.1711.05101
    .. [REDDI2019] Sashank J. Reddi, Satyen Kale, Sanjiv Kumar (2019) On the Convergence of Adam and Beyond. ICLR.
       https://doi.org/10.48550/arXiv.1904.09237

    Parameters
    ----------
    f
        scalar-valued function to be optimized.
    initial_parameters
        sequence (for example list) of parameters to be optimized.
        Note that these parameters will not be changed. Instead, we create a copy and
        leave the initial values untouched.
    max_iterations
        maximum number of iterations.
    learning_rate
        learning rate.
    tolerance_change
        teriminate if the change of objective function is smaller than this tolerance.
    betas
        coefficients used for computing running averages of gradient and its square.
    eps
        term added to the denominator to improve numerical stability.
    weight_decay
        weight decay (L2 penalty if `decoupled_weight_decay` is `False`).
    amsgrad
        whether to use the AMSGrad variant [REDDI2019]_.
    decoupled_weight_decay
        whether to use Adam (default) or AdamW (if set to `True`) [LOS2019]_.
    callback
        function to be called after each iteration. This can be used to monitor the progress of the algorithm.
        If it returns `False`, the algorithm stops at that iteration.

    Returns
    -------
        optimized parameters
    """
    parameters = tuple(p.detach().clone().requires_grad_(True) for p in initial_parameters)

    optim: AdamW | Adam

    if not decoupled_weight_decay:
        optim = Adam(
            params=parameters, lr=learning_rate, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad
        )
    else:
        optim = AdamW(
            params=parameters, lr=learning_rate, betas=betas, eps=eps, weight_decay=weight_decay, amsgrad=amsgrad
        )

    def closure():
        optim.zero_grad()
        (objective,) = f(*parameters)
        objective.backward()

        return objective

    old_objective = float('inf')

    for iteration in range(max_iterations):
        objective = optim.step(closure)

        if tolerance_change > 0:
            objective_ = objective.item()
            change = old_objective - objective_
            if change < tolerance_change:
                break
            old_objective = objective_

        if callback is not None:
            continue_iterations = callback({'solution': parameters, 'iteration_number': iteration})
            if continue_iterations is False:
                break

    parameters = tuple(p.detach() for p in parameters)
    return parameters
